from __future__ import annotations

import enum
import operator
import re
from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable, FrozenSet

from .version import __version__  # noqa: F401

noop = lambda x: x


class ParseError(RuntimeError):
    def __init__(self, expected, stream, index):
        self.expected = expected
        self.stream = stream
        self.index = index

    def line_info(self):
        if isinstance(self.stream, str):
            return f"{self.index.line}:{self.index.column}"
        else:
            return str(self.index.offset)

    def __str__(self):
        expected_list = sorted(repr(e) for e in self.expected)

        if len(expected_list) == 1:
            return f"expected {expected_list[0]} at {self.line_info()}"
        else:
            return f"expected one of {', '.join(expected_list)} at {self.line_info()}"


@dataclass(frozen=True)
class Position:
    offset: int
    line: int
    column: int


@dataclass(frozen=True)
class Result:
    status: bool
    index: Position
    value: Any
    furthest: Position
    expected: FrozenSet[str]

    @staticmethod
    def success(index, value):
        return Result(True, index, value, Position(-1, -1, -1), frozenset())

    @staticmethod
    def failure(index, expected):
        return Result(False, Position(-1, -1, -1), None, index, frozenset([expected]))

    # collect the furthest failure from self and other
    def aggregate(self, other):
        if not other:
            return self

        if self.furthest.offset > other.furthest.offset:
            return self
        elif self.furthest.offset == other.furthest.offset:
            # if we both have the same failure index, we combine the expected messages.
            return Result(
                self.status,
                self.index,
                self.value,
                self.furthest,
                self.expected | other.expected,
            )
        else:
            return Result(
                self.status, self.index, self.value, other.furthest, other.expected
            )


class Parser:
    """
    A Parser is an object that wraps a function whose arguments are
    a string to be parsed and the index on which to begin parsing.
    The function should return either Result.success(next_index, value),
    where the next index is where to continue the parse and the value is
    the yielded value, or Result.failure(index, expected), where expected
    is a string indicating what was expected, and the index is the index
    of the failure.
    """

    def __init__(self, wrapped_fn: Callable[[str | bytes | list, Position], Result]):
        """
        Creates a new Parser from a function that takes a stream
        and returns a Result.
        """
        self.wrapped_fn = wrapped_fn

    def __call__(self, stream: str | bytes | list, index: Position):
        return self.wrapped_fn(stream, index)

    def parse(self, stream: str | bytes | list) -> Any:
        """Parses a string or list of tokens and returns the result or raise a ParseError."""
        (result, _) = (self << eof).parse_partial(stream)
        return result

    def parse_partial(
        self, stream: str | bytes | list
    ) -> tuple[Any, str | bytes | list]:
        """
        Parses the longest possible prefix of a given string.
        Returns a tuple of the result and the unparsed remainder,
        or raises ParseError
        """
        result = self(
            stream,
            Position(0, 0, 0) if isinstance(stream, str) else Position(0, -1, -1),
        )

        if result.status:
            return (result.value, stream[result.index.offset :])
        else:
            raise ParseError(result.expected, stream, result.furthest)

    def bind(self, bind_fn):
        @Parser
        def bound_parser(stream, index):
            result = self(stream, index)

            if result.status:
                next_parser = bind_fn(result.value)
                return next_parser(stream, result.index).aggregate(result)
            else:
                return result

        return bound_parser

    def map(self, map_function: Callable) -> Parser:
        """
        Returns a parser that transforms the produced value of the initial parser with map_function.
        """
        return self.bind(lambda res: success(map_function(res)))

    def combine(self, combine_fn: Callable) -> Parser:
        """
        Returns a parser that transforms the produced values of the initial parser
        with ``combine_fn``, passing the arguments using ``*args`` syntax.

        The initial parser should return a list/sequence of parse results.
        """
        return self.bind(lambda res: success(combine_fn(*res)))

    def combine_dict(self, combine_fn: Callable) -> Parser:
        """
        Returns a parser that transforms the value produced by the initial parser
        using the supplied function/callable, passing the arguments using the
        ``**kwargs`` syntax.

        The value produced by the initial parser must be a mapping/dictionary from
        names to values, or a list of two-tuples, or something else that can be
        passed to the ``dict`` constructor.

        If ``None`` is present as a key in the dictionary it will be removed
        before passing to ``fn``, as will all keys starting with ``_``.
        """
        return self.bind(
            lambda res: success(
                combine_fn(
                    **{
                        k: v
                        for k, v in dict(res).items()
                        if k is not None
                        and not (isinstance(k, str) and k.startswith("_"))
                    }
                )
            )
        )

    def concat(self) -> Parser:
        """
        Returns a parser that concatenates together (as a string) the previously
        produced values.
        """
        return self.map("".join)

    def then(self, other: Parser) -> Parser:
        """
        Returns a parser which, if the initial parser succeeds, will
        continue parsing with ``other``. This will produce the
        value produced by ``other``.

        """
        return seq(self, other).combine(lambda left, right: right)

    def skip(self, other: Parser) -> Parser:
        """
        Returns a parser which, if the initial parser succeeds, will
        continue parsing with ``other``. It will produce the
        value produced by the initial parser.
        """
        return seq(self, other).combine(lambda left, right: left)

    def result(self, value: Any) -> Parser:
        """
        Returns a parser that, if the initial parser succeeds, always produces
        the passed in ``value``.
        """
        return self >> success(value)

    def many(self) -> Parser:
        """
        Returns a parser that expects the initial parser 0 or more times, and
        produces a list of the results.
        """
        return self.times(0, float("inf"))

    def times(self, min: int, max: int = None) -> Parser:
        """
        Returns a parser that expects the initial parser at least ``min`` times,
        and at most ``max`` times, and produces a list of the results. If only one
        argument is given, the parser is expected exactly that number of times.
        """
        if max is None:
            max = min

        @Parser
        def times_parser(stream, index):
            values = []
            times = 0
            result = None

            while times < max:
                result = self(stream, index).aggregate(result)
                if result.status:
                    values.append(result.value)
                    index = result.index
                    times += 1
                elif times >= min:
                    break
                else:
                    return result

            return Result.success(index, values).aggregate(result)

        return times_parser

    def at_most(self, n: int) -> Parser:
        """
        Returns a parser that expects the initial parser at most ``n`` times, and
        produces a list of the results.
        """
        return self.times(0, n)

    def at_least(self, n: int) -> Parser:
        """
        Returns a parser that expects the initial parser at least ``n`` times, and
        produces a list of the results.
        """
        return self.times(n) + self.many()

    def optional(self, default: Any = None) -> Parser:
        """
        Returns a parser that expects the initial parser zero or once, and maps
        the result to a given default value in the case of no match. If no default
        value is given, ``None`` is used.
        """
        return self.times(0, 1).map(lambda v: v[0] if v else default)

    def until(
        self,
        other: Parser,
        min: int = 0,
        max: int = float("inf"),
        consume_other: bool = False,
    ) -> Parser:
        """
        Returns a parser that expects the initial parser followed by ``other``.
        The initial parser is expected at least ``min`` times and at most ``max`` times.
        By default, it does not consume ``other`` and it produces a list of the
        results excluding ``other``. If ``consume_other`` is ``True`` then
        ``other`` is consumed and its result is included in the list of results.
        """

        @Parser
        def until_parser(stream, index):
            values = []
            times = 0
            while True:
                # try parser first
                res = other(stream, index)
                if res.status and times >= min:
                    if consume_other:
                        # consume other
                        values.append(res.value)
                        index = res.index
                    return Result.success(index, values)

                # exceeded max?
                if times >= max:
                    # return failure, it matched parser more than max times
                    return Result.failure(index, f"at most {max} items")

                # failed, try parser
                result = self(stream, index)
                if result.status:
                    # consume
                    values.append(result.value)
                    index = result.index
                    times += 1
                elif times >= min:
                    # return failure, parser is not followed by other
                    return Result.failure(index, "did not find other parser")
                else:
                    # return failure, it did not match parser at least min times
                    return Result.failure(
                        index, f"at least {min} items; got {times} item(s)"
                    )

        return until_parser

    def sep_by(self, sep: Parser, *, min: int = 0, max: int = float("inf")) -> Parser:
        """
        Returns a new parser that repeats the initial parser and
        collects the results in a list. Between each item, the ``sep`` parser
        is run (and its return value is discarded). By default it
        repeats with no limit, but minimum and maximum values can be supplied.
        """
        zero_times = success([])
        if max == 0:
            return zero_times
        res = self.times(1) + (sep >> self).times(min - 1, max - 1)
        if min == 0:
            res |= zero_times
        return res

    def desc(self, description: str) -> Parser:
        """
        Returns a new parser with a description added, which is used in the error message
        if parsing fails.
        """

        @Parser
        def desc_parser(stream, index):
            result = self(stream, index)
            if result.status:
                return result
            else:
                return Result.failure(index, description)

        return desc_parser

    def mark(self) -> Parser:
        """
        Returns a parser that wraps the initial parser's result in a value
        containing column and line information of the match, as well as the
        original value. The new value is a 3-tuple:

        ((start_row, start_column),
         original_value,
         (end_row, end_column))
        """

        @generate
        def marked():
            start = yield line_info
            body = yield self
            end = yield line_info
            return (start, body, end)

        return marked

    def tag(self, name: str) -> Parser:
        """
        Returns a parser that wraps the produced value of the initial parser in a
        2 tuple containing ``(name, value)``. This provides a very simple way to
        label parsed components
        """
        return self.map(lambda v: (name, v))

    def should_fail(self, description: str) -> Parser:
        """
        Returns a parser that fails when the initial parser succeeds, and succeeds
        when the initial parser fails (consuming no input). A description must
        be passed which is used in parse failure messages.

        This is essentially a negative lookahead
        """

        @Parser
        def fail_parser(stream, index):
            res = self(stream, index)
            if res.status:
                return Result.failure(index, description)
            return Result.success(index, res)

        return fail_parser

    def __add__(self, other: Parser) -> Parser:
        return seq(self, other).combine(operator.add)

    def __mul__(self, other: Parser) -> Parser:
        if isinstance(other, range):
            return self.times(other.start, other.stop - 1)
        return self.times(other)

    def __or__(self, other: Parser) -> Parser:
        return alt(self, other)

    # haskelley operators, for fun #

    # >>
    def __rshift__(self, other: Parser) -> Parser:
        return self.then(other)

    # <<
    def __lshift__(self, other: Parser) -> Parser:
        return self.skip(other)


def alt(*parsers: Parser) -> Parser:
    """
    Creates a parser from the passed in argument list of alternative
    parsers, which are tried in order, moving to the next one if the
    current one fails.
    """
    if not parsers:
        return fail("<empty alt>")

    @Parser
    def alt_parser(stream, index):
        result = None
        for parser in parsers:
            result = parser(stream, index).aggregate(result)
            if result.status:
                return result

        return result

    return alt_parser


def seq(*parsers: Parser, **kw_parsers: Parser) -> Parser:
    """
    Takes a list of parsers, runs them in order,
    and collects their individuals results in a list,
    or in a dictionary if you pass them as keyword arguments.
    """
    if not parsers and not kw_parsers:
        return success([])

    if parsers and kw_parsers:
        raise ValueError(
            "Use either positional arguments or keyword arguments with seq, not both"
        )

    if parsers:

        @Parser
        def seq_parser(stream, index):
            result = None
            values = []
            for parser in parsers:
                result = parser(stream, index).aggregate(result)
                if not result.status:
                    return result
                index = result.index
                values.append(result.value)
            return Result.success(index, values).aggregate(result)

        return seq_parser
    else:

        @Parser
        def seq_kwarg_parser(stream, index):
            result = None
            values = {}
            for name, parser in kw_parsers.items():
                result = parser(stream, index).aggregate(result)
                if not result.status:
                    return result
                index = result.index
                values[name] = result.value
            return Result.success(index, values).aggregate(result)

        return seq_kwarg_parser


def generate(fn) -> Parser:
    """
    Creates a parser from a generator function
    """
    if isinstance(fn, str):
        return lambda f: generate(f).desc(fn)

    @Parser
    @wraps(fn)
    def generated(stream, index):
        # start up the generator
        iterator = fn()

        result = None
        value = None
        try:
            while True:
                next_parser = iterator.send(value)
                result = next_parser(stream, index).aggregate(result)
                if not result.status:
                    return result
                value = result.value
                index = result.index
        except StopIteration as stop:
            returnVal = stop.value
            if isinstance(returnVal, Parser):
                return returnVal(stream, index).aggregate(result)

            return Result.success(index, returnVal).aggregate(result)

    return generated


index = Parser(lambda _, index: Result.success(index, index.offset))
line_info = Parser(lambda _, index: Result.success(index, (index.line, index.column)))


def success(value: Any) -> Parser:
    """
    Returns a parser that does not consume any of the stream, but
    produces ``value``.
    """
    return Parser(lambda _, index: Result.success(index, value))


def fail(expected: str) -> Parser:
    """
    Returns a parser that always fails with the provided error message.
    """
    return Parser(lambda _, index: Result.failure(index, expected))


def make_index_update(consumed: str) -> Callable[[Position], Position]:
    slen = len(consumed)
    line_count = consumed.count("\n")
    last_nl = consumed.rfind("\n")
    return lambda index: Position(
        offset=index.offset + slen,
        line=index.line + line_count,
        column=slen - (last_nl + 1) if last_nl >= 0 else index.column + slen,
    )


def string(expected_string: str, transform: Callable[[str], str] = noop) -> Parser:
    """
    Returns a parser that expects the ``expected_string`` and produces
    that string value.

    Optionally, a transform function can be passed, which will be used on both
    the expected string and tested string.
    """

    slen = len(expected_string)
    transformed_s = transform(expected_string)
    index_update = make_index_update(expected_string)

    @Parser
    def string_parser(stream, index):
        if transform(stream[index.offset : index.offset + slen]) == transformed_s:
            return Result.success(index_update(index), expected_string)
        else:
            return Result.failure(index, expected_string)

    return string_parser


def regex(exp: str, flags=0, group: int | str | tuple = 0) -> Parser:
    """
    Returns a parser that expects the given ``exp``, and produces the
    matched string. ``exp`` can be a compiled regular expression, or a
    string which will be compiled with the given ``flags``.

    Optionally, accepts ``group``, which is passed to re.Match.group
    https://docs.python.org/3/library/re.html#re.Match.group> to
    return the text from a capturing group in the regex instead of the
    entire match.
    """

    if isinstance(exp, (str, bytes)):
        exp = re.compile(exp, flags)
    if isinstance(group, (str, int)):
        group = (group,)

    @Parser
    def regex_parser(stream, index):
        match = exp.match(stream, index.offset)
        if match:
            index = (
                make_index_update(stream[match.start() : match.end()])(index)
                if isinstance(stream, str)
                else Position(match.end(), -1, -1)
            )
            return Result.success(index, match.group(*group))
        else:
            return Result.failure(index, exp.pattern)

    return regex_parser


def test_item(func: Callable[..., bool], description: str) -> Parser:
    """
    Returns a parser that tests a single item from the list of items being
    consumed, using the callable ``func``. If ``func`` returns ``True``, the
    parse succeeds, otherwise the parse fails with the description
    ``description``.
    """

    @Parser
    def test_item_parser(stream, index):
        if index.offset < len(stream):
            if isinstance(stream, bytes):
                # Subscripting bytes with `[index]` instead of
                # `[index:index + 1]` returns an int
                item = stream[index.offset : index.offset + 1]
            else:
                item = stream[index.offset]
            if func(item):
                if isinstance(stream, str):
                    index = make_index_update(item)(index)
                else:
                    index = Position(index.offset + 1, index.line, index.column)
                return Result.success(index, item)
        return Result.failure(index, description)

    return test_item_parser


def test_char(func: Callable[..., bool], description: str) -> Parser:
    """
    Returns a parser that tests a single character with the callable
    ``func``. If ``func`` returns ``True``, the parse succeeds, otherwise
    the parse fails with the description ``description``.
    """
    # Implementation is identical to test_item
    return test_item(func, description)


def match_item(item: Any, description: str = None) -> Parser:
    """
    Returns a parser that tests the next item (or character) from the stream (or
    string) for equality against the provided item. Optionally a string
    description can be passed.
    """

    if description is None:
        description = str(item)
    return test_item(lambda i: item == i, description)


def string_from(*strings: str, transform: Callable[[str], str] = noop):
    """
    Accepts a sequence of strings as positional arguments, and returns a parser
    that matches and returns one string from the list. The list is first sorted
    in descending length order, so that overlapping strings are handled correctly
    by checking the longest one first.
    """
    # Sort longest first, so that overlapping options work correctly
    return alt(*(string(s, transform) for s in sorted(strings, key=len, reverse=True)))


def char_from(string: str | bytes):
    """
    Accepts a string and returns a parser that matches and returns one character
    from the string.
    """
    if isinstance(string, bytes):
        return test_char(lambda c: c in string, b"[" + string + b"]")
    else:
        return test_char(lambda c: c in string, "[" + string + "]")


def peek(parser: Parser) -> Parser:
    """
    Returns a lookahead parser that parses the input stream without consuming
    chars.
    """

    @Parser
    def peek_parser(stream, index):
        result = parser(stream, index)
        if result.status:
            return Result.success(index, result.value)
        else:
            return result

    return peek_parser


any_char = test_char(lambda c: True, "any character")

whitespace = regex(r"\s+")

letter = test_char(lambda c: c.isalpha(), "a letter")

digit = test_char(lambda c: c.isdigit(), "a digit")

decimal_digit = char_from("0123456789")


@Parser
def eof(stream, index):
    """
    A parser that only succeeds if the end of the stream has been reached.
    """

    if index.offset >= len(stream):
        return Result.success(index, None)
    else:
        return Result.failure(index, "EOF")


def from_enum(enum_cls: type[enum.Enum], transform=noop) -> Parser:
    """
    Given a class that is an enum.Enum class
    https://docs.python.org/3/library/enum.html , returns a parser that
    will parse the values (or the string representations of the values)
    and return the corresponding enum item.
    """

    items = sorted(
        ((str(enum_item.value), enum_item) for enum_item in enum_cls),
        key=lambda t: len(t[0]),
        reverse=True,
    )
    return alt(
        *(
            string(value, transform=transform).result(enum_item)
            for value, enum_item in items
        )
    )


class forward_declaration(Parser):
    """
    An empty parser that can be used as a forward declaration,
    especially for parsers that need to be defined recursively.

    You must use `.become(parser)` before using.
    """

    def __init__(self):
        pass

    def _raise_error(self, *args, **kwargs):
        raise ValueError(
            "You must use 'become' before attempting to call `parse` or `parse_partial`"
        )

    parse = _raise_error
    parse_partial = _raise_error

    def become(self, other: Parser):
        """
        Take on the behavior of the given parser.
        """
        self.__dict__ = other.__dict__
        self.__class__ = other.__class__



import sys
from typing import Any
from typing import Dict
from typing import Optional

import requests
from attr import define
from attr import Factory
from attr import field

from semgrep.error import FINDINGS_EXIT_CODE
from semgrep.error import OK_EXIT_CODE
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)


@define
class ErrorHandler:
    """
    Send scan status to the fail-open endpoint
    """

    suppress_errors: bool = field(default=False)
    payload: Dict[str, Any] = Factory(dict)

    def configure(self, suppress_errors: Optional[bool] = False) -> None:
        """
        Configures whether to always or never send fail-open status.

        :param suppress_errors: The value of the --suppress-errors option
        """
        if suppress_errors:
            self.suppress_errors = suppress_errors

    def push_request(self, method: str, url: str, **kwargs: Any) -> None:
        self.payload = {"method": method, "url": url}

    def append_request(self, **kwargs: Any) -> None:
        self.payload = {**self.payload, **kwargs}

    def pop_request(self) -> None:
        self.payload = {}

    @property
    def is_enabled(self) -> bool:
        """
        Returns whether scan status should be sent.
        """
        return self.suppress_errors

    def send(self, exit_code: int) -> int:
        """
        Send scan status to the fail-open server.

        Status will only be emitted if is_enabled is True
        """
        from semgrep.state import get_state  # avoiding circular import
        from semgrep.app import auth  # avoiding circular import

        state = get_state()

        if (
            not self.is_enabled
            or exit_code == OK_EXIT_CODE
            or exit_code == FINDINGS_EXIT_CODE
        ):
            return exit_code

        import traceback

        logger.debug(
            f"Sending to fail-open endpoint {state.env.fail_open_url} since fail-open is configured to {self.suppress_errors}"
        )

        token = auth.get_token()
        headers = {
            "User-Agent": str(state.app_session.user_agent),
            "Authorization": f"Bearer {token or ''}",
        }
        if sys.exc_info()[0] is not None:
            self.payload["error"] = traceback.format_exc()

        try:
            requests.post(
                state.env.fail_open_url, headers=headers, json=self.payload, timeout=3
            )
        except Exception as e:
            logger.debug(f"Error sending to fail-open endpoint: {e}")

        return 0



"""
Custom parser and comparator for Maven versions
Based on https://docs.oracle.com/middleware/1212/core/MAVEN/maven_version.htm#MAVEN400
"""
import re
from dataclasses import dataclass
from typing import Tuple
from typing import Union

from semdep.external.parsy import ParseError
from semdep.external.parsy import string
from semdep.parsers.util import any_str
from semdep.parsers.util import integer
from semdep.parsers.util import pair
from semdep.parsers.util import triple
from semgrep.error import SemgrepError


@dataclass
class VersionCore:
    major: int
    minor: int
    incrementals: Tuple[int, ...]


parse_version_core = triple(
    integer,
    string(".") >> integer,
    (string(".") >> integer).many(),
).map(lambda x: VersionCore(x[0], x[1], tuple(x[2])))


def normalize_incrementals(
    first: Tuple[int, ...], second: Tuple[int, ...]
) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
    if len(first) > len(second):
        return first, second + (0,) * (len(first) - len(second))
    elif len(second) > len(first):
        return first + (0,) * (len(second) - len(first)), second
    else:
        return first, second


def compare_version_core(first: VersionCore, second: VersionCore) -> int:
    if first.major != second.major:
        return first.major - second.major
    elif first.minor != second.minor:
        return first.minor - second.minor
    for inc1, inc2 in zip(
        *normalize_incrementals(first.incrementals, second.incrementals)
    ):
        if inc1 != inc2:
            return inc1 - inc2
    return 0


@dataclass
class ParsedMavenVersion:
    core: VersionCore
    qualifier: str
    raw_version: str


MavenVersion = Union[ParsedMavenVersion, str]


def parse_maven_version(version: str) -> MavenVersion:
    try:
        m = pair(parse_version_core, any_str).parse(version)
    except ParseError:
        # "If you do not follow Maven versioning standards in your project versioning scheme,
        # then for version comparison, Maven interprets the entire version as a simple string."
        return version
    return ParsedMavenVersion(core=m[0], qualifier=m[1], raw_version=version)


def cmp_str(x: str, y: str) -> int:
    if x == y:
        return 0
    elif x > y:
        return 1
    else:
        return -1


def cmp_maven_versions(first: MavenVersion, second: MavenVersion) -> int:
    """
    Return less than 0 if first < second
    0 if ==
    greater than 0 if first > second
    """
    if isinstance(first, ParsedMavenVersion) and isinstance(second, ParsedMavenVersion):
        core_cmp = compare_version_core(first.core, second.core)
        if core_cmp != 0:
            return core_cmp
        # "All versions with a qualifier are older than the same version without a qualifier (release version)."
        elif first.qualifier == "" and second.qualifier != "":
            return 1
        elif second.qualifier == "" and first.qualifier != "":
            return -1

        # "Maven treats the SNAPSHOT qualifier differently from all others.
        # If a version number is followed by -SNAPSHOT, then Maven considers it
        # the "as-yet-unreleased" version of the associated MajorVersion, MinorVersion, or IncrementalVersion."
        elif first.qualifier == "-SNAPSHOT" and second.qualifier != "-SNAPSHOT":
            return 1
        elif second.qualifier == "-SNAPSHOT" and first.qualifier != "-SNAPSHOT":
            return -1
        else:
            return cmp_str(first.qualifier, second.qualifier)
    else:
        if isinstance(first, ParsedMavenVersion) and isinstance(second, str):
            first_raw = first.raw_version
            second_raw = second
        elif isinstance(second, ParsedMavenVersion) and isinstance(first, str):
            first_raw = first
            second_raw = second.raw_version
        else:
            assert isinstance(first, str)
            assert isinstance(second, str)
            first_raw = first
            second_raw = second
        return cmp_str(first_raw, second_raw)


def compare_maven_specifier(specifier: str, version: str) -> bool:
    """
    Returns if version satisfies specifier requirement

    i.e. specifier: '< 1.0.0', version: 0.1.0 returns true

    See https://docs.oracle.com/middleware/1212/core/MAVEN/maven_version.htm#MAVEN400
    """
    specifier_regex = re.compile(
        r"""(?P<operator>(==|!=|<=|>=|<|>))\s*(?P<version>.*)"""
    )
    matched = specifier_regex.match(specifier)
    if not matched:
        raise SemgrepError(
            f"unknown package version comparison expression: {specifier}"
        )
    operator = matched.group("operator")
    specifier_version = parse_maven_version(matched.group("version"))
    parsed_version = parse_maven_version(version)

    if operator == "==":
        return cmp_maven_versions(parsed_version, specifier_version) == 0
    elif operator == "!=":
        return cmp_maven_versions(parsed_version, specifier_version) != 0
    elif operator == "<=":
        return cmp_maven_versions(parsed_version, specifier_version) <= 0
    elif operator == "<":
        return cmp_maven_versions(parsed_version, specifier_version) < 0
    elif operator == ">=":
        return cmp_maven_versions(parsed_version, specifier_version) >= 0
    elif operator == ">":
        return cmp_maven_versions(parsed_version, specifier_version) > 0
    else:
        raise SemgrepError(
            f"unknown package version comparison expression: {specifier}"
        )



"""
Parsers for yarn.lock versions 1 and 2/3
Version 1 parser based on
https://www.arahansen.com/the-ultimate-guide-to-yarn-lock-lockfiles/
https://classic.yarnpkg.com/lang/en/docs/yarn-lock/

Version 2/3 parser based on looking at examples on github, I could not find any documentation
Here are the Yarn 2/3 docs: https://yarnpkg.com/
"""
from pathlib import Path
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TypeVar

from semdep.external.parsy import any_char
from semdep.external.parsy import Parser
from semdep.external.parsy import peek
from semdep.external.parsy import regex
from semdep.external.parsy import string
from semdep.external.parsy import success
from semdep.parsers.util import consume_line
from semdep.parsers.util import extract_npm_lockfile_hash
from semdep.parsers.util import json_doc
from semdep.parsers.util import mark_line
from semdep.parsers.util import pair
from semdep.parsers.util import ParserName
from semdep.parsers.util import quoted
from semdep.parsers.util import safe_path_parse
from semdep.parsers.util import transitivity
from semdep.parsers.util import upto
from semgrep.semgrep_interfaces.semgrep_output_v1 import Ecosystem
from semgrep.semgrep_interfaces.semgrep_output_v1 import FoundDependency
from semgrep.semgrep_interfaces.semgrep_output_v1 import Npm
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)

A = TypeVar("A")

# The initial line of a yarn version 1 dependency, lists the constraints that lead to this package
# Examples:
# "@ampproject/remapping@^2.0.0":
# bad-lib@0.0.8:
# my-package-without-version-constraint:
# "filedep@file:../../correct/path/filedep":
# "bats@https://github.com/bats-core/bats-core#master":
part1 = regex('"?@?([^@:]*)', flags=0, group=1)
part2 = regex('@?([^:,"]*(:?(?!\n)[^:,"]*)*)"?', flags=0, group=1)
source1 = pair(part1, part2)

# Examples:
# "@ampproject/remapping@^2.0.0", "@ampproject/remapping@^3.1.0"
# bad-lib@0.0.8, bad-lib@^0.0.4
multi_source1 = source1.sep_by(string(", "))

# A key value pair. These can be a name followed by a nested list, but the only data we care about is in outermost list
# This is why we produce None if the line is preceeded by more than 2 spaces, or if it ends in a colon
# Examples:
#   version "2.1.1"
#   integrity sha512-Aolwjd7HSC2PyY0fDj/wA/EimQT4HfEnFYNp5s9CQlrdhyvWTtvZ5YzrUPu6R6/1jKiUlxu8bUhkdSnKHNAHMA==
#   dependencies:
key_value1: "Parser[Optional[Tuple[str,str]]]" = (
    string(" ")
    .many()
    .bind(
        lambda spaces: consume_line
        if len(spaces) != 2
        else upto(" ", ":").bind(
            lambda key: peek(any_char).bind(
                lambda next: consume_line
                if next == ":"
                else string(" ")
                >> upto("\n").bind(lambda value: success((key, value.strip('"'))))  # type: ignore
                # mypy seemingly cannot figure out that this function returns an optional
            )
        )
    )
)

# A full spec of a dependency
# Examples:
# "@ampproject/remapping@^2.0.0":
#   version "2.1.1"
#   resolved "https://registry.npmjs.org/@ampproject/remapping/-/remapping-2.1.1.tgz"
#   integrity sha512-Aolwjd7HSC2PyY0fDj/wA/EimQT4HfEnFYNp5s9CQlrdhyvWTtvZ5YzrUPu6R6/1jKiUlxu8bUhkdSnKHNAHMA==
#   dependencies:
#     "@jridgewell/trace-mapping" "^0.3.0"
yarn_dep1 = mark_line(
    pair(
        multi_source1 << string(":\n"),
        key_value1.sep_by(string("\n")).map(lambda xs: {x[0]: x[1] for x in xs if x}),
    )
)

YARN1_PREFIX = """\
# THIS IS AN AUTOGENERATED FILE. DO NOT EDIT THIS FILE DIRECTLY.
# yarn lockfile v1

"""

yarn1 = (
    string(YARN1_PREFIX)
    >> string("\n").optional()
    >> yarn_dep1.sep_by(string("\n\n"))
    << string("\n").optional()
)


# The yarn version 2/3 parser is set up equivalently, with slight differences in sub-parsers


def remove_npm_prefix(s: str) -> str:
    if s.startswith("npm:"):
        return s[4:]
    else:
        return s


# Examples:
# @ampproject/remapping@npm:^2.0.0
# @my-scope/my-first-package@my-scope/my-first-package#commit=0b824c650d3a03444dbcf2b27a5f3566f6e41358
# my-third-package@https://github.com/my-org/my-third-package#everything
# my-package@file:../../deps/my-local-package::locator=my-project%40workspace%3A.
# resolve@patch:resolve@^1.1.7#~builtin<compat/resolve>
source2 = pair(
    string("@").optional("") + upto("@", consume_other=True),
    # We remove the "npm:" prefix, because in a package.json, the version constraint will appear without it
    # e.g. "^1.0.0" in package.json becomes "npm:^1.0.0" in yarn.lock
    # However, prefix like "file:" *do* appear in package.json, so they aren't removed
    upto('"', ",").map(remove_npm_prefix),
)

# Examples:
# "@apidevtools/json-schema-ref-parser@npm:9.0.9"
# "@babel/generator@npm:^7.12.11, @babel/generator@npm:^7.12.5, @babel/generator@npm:^7.18.10"
multi_source2 = quoted(source2.sep_by(string(", ")))

# Examples:
#   version: 7.18.10
#   resolution: "@babel/generator@npm:7.18.10"
#   dependencies:
key_value2: "Parser[Optional[Tuple[str,str]]]" = (
    string(" ")
    .many()
    .bind(
        lambda spaces: consume_line
        if len(spaces) != 2
        else upto(":").bind(
            lambda key: string(":")
            >> peek(any_char).bind(
                lambda next: success(None)
                if next == "\n"
                else string(" ")
                >> upto("\n").bind(lambda value: success((key, value.strip('"'))))  # type: ignore
                # mypy seemingly cannot figure out that this function returns an optional
            )
        )
    )
)

# Examples:
# "@babel/generator@npm:^7.17.0, @babel/generator@npm:^7.7.2":
#   version: 7.17.0
#   resolution: "@babel/generator@npm:7.17.0"
#   dependencies:
#     "@babel/types": ^7.17.0
#     jsesc: ^2.5.1
#     source-map: ^0.5.0
#   checksum: 2987dbebb484727a227f1ce3db90810320986cfb3ffd23e6d1d87f75bbd8e7871b5bc44252822d4d5f048a2d872a5702b2a9bf7bab7e07f087d7f306f0ea6c0a
#   languageName: node
#   linkType: hard
yarn_dep2 = mark_line(
    pair(
        multi_source2 << string(":\n"),
        key_value2.sep_by(string("\n")).map(lambda xs: {x[0]: x[1] for x in xs if x}),
    )
)

YARN2_PREFIX = """\
# This file is generated by running "yarn install" inside your project.
# Manual changes might be lost - proceed with caution!
"""
YARN2_METADATA_REGEX = """\

__metadata:
  version: \\d+
  cacheKey: \\d+
"""
yarn2 = (
    string(YARN2_PREFIX)
    >> regex(YARN2_METADATA_REGEX).optional()
    >> string("\n").optional()
    >> yarn_dep2.sep_by(string("\n\n"))
    << string("\n").optional()
)


def get_manifest_deps(manifest_path: Optional[Path]) -> Optional[Set[Tuple[str, str]]]:
    """
    Extract a set of constraints from a package.json file
    """
    if not manifest_path:
        return None
    json_opt = safe_path_parse(manifest_path, json_doc, ParserName.jsondoc)
    if not json_opt:
        return None
    json = json_opt.as_dict()
    deps = json.get("dependencies")
    if not deps:
        return set()
    return {(x[0], x[1].as_str()) for x in deps.as_dict().items()}


def remove_trailing_octothorpe(s: Optional[str]) -> Optional[str]:
    if s is None:
        return None
    else:
        return "#".join(s.split("#")[:-1]) if "#" in s else s


def parse_yarn(
    lockfile_path: Path, manifest_path: Optional[Path]
) -> List[FoundDependency]:
    with open(lockfile_path) as f:
        lockfile_text = f.read()
    manifest_deps = get_manifest_deps(manifest_path)
    yarn_version = 1 if lockfile_text.startswith(YARN1_PREFIX) else 2
    parser = yarn1 if yarn_version == 1 else yarn2
    parser_name = ParserName.yarn_1 if yarn_version == 1 else ParserName.yarn_2
    deps = safe_path_parse(lockfile_path, parser, parser_name)
    if not deps:
        return []
    output = []
    for line_number, (sources, fields) in deps:
        if len(sources) < 1:
            continue
        if "version" not in fields:
            continue
        if yarn_version == 1:
            allowed_hashes = extract_npm_lockfile_hash(fields.get("integrity"))
        else:
            checksum = fields.get("checksum")
            allowed_hashes = {"sha512": [checksum]} if checksum else {}
        resolved_url = fields.get("resolved")
        output.append(
            FoundDependency(
                package=sources[0][0],
                version=fields["version"],
                ecosystem=Ecosystem(Npm()),
                allowed_hashes=allowed_hashes,
                resolved_url=remove_trailing_octothorpe(resolved_url),
                transitivity=transitivity(manifest_deps, sources),
                line_number=line_number,
            )
        )
    return output



"""
This file encapsulates classes necessary in parsing semgrep-core
json output into a typed object

The precise type of the response from semgrep-core is specified in
https://github.com/returntocorp/semgrep/blob/develop/interfaces/Output_from_core.atd
"""
import dataclasses
from dataclasses import replace
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import semgrep.output_from_core as core
import semgrep.semgrep_interfaces.semgrep_output_v1 as out
import semgrep.util as util
from semgrep.error import FATAL_EXIT_CODE
from semgrep.error import Level
from semgrep.error import SemgrepCoreError
from semgrep.error import SemgrepError
from semgrep.error import TARGET_PARSE_FAILURE_EXIT_CODE
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch
from semgrep.rule_match import RuleMatchSet
from semgrep.types import JsonObject
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)


def _core_location_to_error_span(location: core.Location) -> out.ErrorSpan:
    return out.ErrorSpan(
        file=location.path,
        start=out.PositionBis(line=location.start.line, col=location.start.col),
        end=out.PositionBis(line=location.end.line, col=location.end.col),
    )


def core_error_to_semgrep_error(err: core.CoreError) -> SemgrepCoreError:
    # Hackily convert the level string to Semgrep expectations
    level_str = err.severity.kind
    if level_str.upper() == "WARNING":
        level_str = "WARN"
    if level_str.upper() == "ERROR_":
        level_str = "ERROR"
    level = Level[level_str.upper()]

    spans: Optional[List[out.ErrorSpan]] = None
    if isinstance(err.error_type.value, core.PatternParseError):
        yaml_path = err.error_type.value.value[::-1]
        error_span = _core_location_to_error_span(err.location)
        config_start = out.PositionBis(line=0, col=1)
        config_end = out.PositionBis(
            line=err.location.end.line - err.location.start.line,
            col=err.location.end.col - err.location.start.col + 1,
        )
        spans = [
            dataclasses.replace(
                error_span,
                config_start=config_start,
                config_end=config_end,
                config_path=yaml_path,
            )
        ]
    elif isinstance(err.error_type.value, core.PartialParsing):
        # The spans for PartialParsing errors are contained in the "error_type" object
        spans = [
            _core_location_to_error_span(location)
            for location in err.error_type.value.value
        ]

    # TODO benchmarking code relies on error code value right now
    # See https://semgrep.dev/docs/cli-usage/ for meaning of codes
    if (
        isinstance(err.error_type.value, core.ParseError)
        or isinstance(err.error_type.value, core.LexicalError)
        or isinstance(err.error_type.value, core.PartialParsing)
    ):
        code = TARGET_PARSE_FAILURE_EXIT_CODE
        err = replace(err, rule_id=None)  # Rule id not important for parse errors
    elif isinstance(err.error_type.value, core.PatternParseError):
        # TODO This should probably be RULE_PARSE_FAILURE_EXIT_CODE
        # but we have been exiting with FATAL_EXIT_CODE, so we need
        # to be deliberate about changing it
        code = FATAL_EXIT_CODE
    else:
        code = FATAL_EXIT_CODE

    return SemgrepCoreError(code, level, spans, err)


def parse_core_output(raw_json: JsonObject) -> core.CoreMatchResults:
    match_results = core.CoreMatchResults.from_json(raw_json)
    if match_results.skipped_targets:
        for skip in match_results.skipped_targets:
            if skip.rule_id:
                rule_info = f"rule {skip.rule_id}"
            else:
                rule_info = "all rules"
            logger.verbose(
                f"skipped '{skip.path}' [{rule_info}]: {skip.reason}: {skip.details}"
            )
    return match_results


def core_matches_to_rule_matches(
    rules: List[Rule], res: core.CoreMatchResults
) -> Dict[Rule, List[RuleMatch]]:
    """
    Convert core_match objects into RuleMatch objects that the rest of the codebase
    interacts with.

    For now assumes that all matches encapsulated by this object are from the same rule
    """
    rule_table = {rule.id: rule for rule in rules}

    def interpolate(
        text: str, metavariables: Dict[str, str], propagated_values: Dict[str, str]
    ) -> str:
        """Interpolates a string with the metavariables contained in it, returning a new string"""

        # Sort by metavariable length to avoid name collisions (eg. $X2 must be handled before $X)
        for metavariable in sorted(metavariables.keys(), key=len, reverse=True):
            text = text.replace(
                "value(" + metavariable + ")", propagated_values[metavariable]
            )
            text = text.replace(metavariable, metavariables[metavariable])

        return text

    def read_metavariables(
        match: core.CoreMatch,
    ) -> Tuple[Dict[str, str], Dict[str, str]]:
        matched_values = {}
        propagated_values = {}

        # open path and ignore non-utf8 bytes. https://stackoverflow.com/a/56441652
        with open(match.location.path.value, errors="replace") as fd:
            for metavariable, metavariable_data in match.extra.metavars.value.items():
                # Offsets are start inclusive and end exclusive
                start_offset = metavariable_data.start.offset
                end_offset = metavariable_data.end.offset

                matched_value = util.read_range(fd, start_offset, end_offset)

                # Use propagated value
                if metavariable_data.propagated_value:
                    propagated_value = (
                        metavariable_data.propagated_value.svalue_abstract_content
                    )
                else:
                    propagated_value = matched_value

                matched_values[metavariable] = matched_value
                propagated_values[metavariable] = propagated_value

        return matched_values, propagated_values

    def convert_to_rule_match(match: core.CoreMatch) -> RuleMatch:
        rule = rule_table[match.rule_id.value]
        matched_values, propagated_values = read_metavariables(match)
        message = interpolate(rule.message, matched_values, propagated_values)
        if match.extra.rendered_fix is not None:
            fix = match.extra.rendered_fix
            logger.debug(f"Using AST-based autofix rendered in semgrep-core: `{fix}`")
        elif rule.fix is not None:
            fix = interpolate(rule.fix, matched_values, propagated_values)
            logger.debug(f"Using text-based autofix rendered in cli: `{fix}`")
        else:
            fix = None
        fix_regex = None

        # this validation for fix_regex code was in autofix.py before
        # TODO: this validation should be done in rule.py when parsing the rule
        if rule.fix_regex:
            regex = rule.fix_regex.get("regex")
            replacement = rule.fix_regex.get("replacement")
            count = rule.fix_regex.get("count")
            if not regex or not replacement:
                raise SemgrepError(
                    "'regex' and 'replacement' values required when using 'fix-regex'"
                )
            if count:
                try:
                    count = int(count)
                except ValueError:
                    raise SemgrepError(
                        "optional 'count' value must be an integer when using 'fix-regex'"
                    )

            fix_regex = out.FixRegex(regex=regex, replacement=replacement, count=count)

        return RuleMatch(
            match=match,
            extra=match.extra.to_json(),
            message=message,
            metadata=rule.metadata,
            severity=rule.severity,
            fix=fix,
            fix_regex=fix_regex,
        )

    # TODO: Dict[core.RuleId, RuleMatchSet]
    findings: Dict[Rule, RuleMatchSet] = {rule: RuleMatchSet(rule) for rule in rules}
    seen_cli_unique_keys: Set[Tuple] = set()
    for match in res.matches:
        rule = rule_table[match.rule_id.value]
        rule_match = convert_to_rule_match(match)
        if rule_match.cli_unique_key in seen_cli_unique_keys:
            continue
        seen_cli_unique_keys.add(rule_match.cli_unique_key)
        findings[rule].add(rule_match)

    # Sort results so as to guarantee the same results across different
    # runs. Results may arrive in a different order due to parallelism
    # (-j option).
    return {rule: sorted(matches) for rule, matches in findings.items()}



# Handle communication of findings / errors to semgrep.app
import json
import os
from collections import Counter
from copy import deepcopy
from pathlib import Path
from typing import Any
from typing import Dict
from typing import FrozenSet
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING
from urllib.parse import urlencode

import click
import requests
from boltons.iterutils import partition

import semgrep.semgrep_interfaces.semgrep_output_v1 as out
from semdep.parsers.util import DependencyParserError
from semgrep.constants import DEFAULT_SEMGREP_APP_CONFIG_URL
from semgrep.constants import RuleSeverity
from semgrep.error import SemgrepError
from semgrep.parsing_data import ParsingData
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatchMap
from semgrep.state import get_state
from semgrep.verbose_logging import getLogger


if TYPE_CHECKING:
    from semgrep.engine import EngineType

logger = getLogger(__name__)


class ScanHandler:
    def __init__(self, dry_run: bool = False, deployment_name: str = "") -> None:
        self._deployment_id: Optional[int] = None
        self._deployment_name: str = deployment_name

        self.scan_id = None
        self.ignore_patterns: List[str] = []
        self._policy_names: List[str] = []
        self._autofix = False
        self._deepsemgrep = False
        self.dry_run = dry_run
        self._dry_run_rules_url: str = ""
        self._skipped_syntactic_ids: List[str] = []
        self._skipped_match_based_ids: List[str] = []
        self._scan_params: str = ""
        self._rules: str = ""

    @property
    def deployment_id(self) -> Optional[int]:
        """
        Separate property for easy of mocking in test
        """
        return self._deployment_id

    @property
    def deployment_name(self) -> str:
        """
        Separate property for easy of mocking in test
        """
        return self._deployment_name

    @property
    def policy_names(self) -> List[str]:
        """
        Separate property for easy of mocking in test
        """
        return self._policy_names

    @property
    def autofix(self) -> bool:
        """
        Separate property for easy of mocking in test
        """
        return self._autofix

    @property
    def deepsemgrep(self) -> bool:
        """
        Separate property for easy of mocking in test
        """
        return self._deepsemgrep

    @property
    def skipped_syntactic_ids(self) -> List[str]:
        """
        Separate property for easy of mocking in test
        """
        return self._skipped_syntactic_ids

    @property
    def skipped_match_based_ids(self) -> List[str]:
        """
        Separate property for easy of mocking in test
        """
        return self._skipped_match_based_ids

    @property
    def scan_params(self) -> str:
        """
        Separate property for easy of mocking in test
        """
        return self._scan_params

    @property
    def rules(self) -> str:
        """
        Separate property for easy of mocking in test
        """
        return self._rules

    def _get_scan_config_from_app(self, url: str) -> Dict[str, Any]:
        state = get_state()
        response = state.app_session.get(url)
        try:
            response.raise_for_status()
        except requests.RequestException:
            raise Exception(
                f"API server at {state.env.semgrep_url} returned this error: {response.text}"
            )

        body = response.json()

        if not isinstance(body, dict):
            raise Exception(
                f"API server at {state.env.semgrep_url} returned type '{type(body)}'. Expected a dictionary."
            )

        return body

    def fetch_and_init_scan_config(self, meta: Dict[str, Any]) -> None:
        """
        Get configurations for scan
        """
        state = get_state()

        self._scan_params = urlencode(
            {
                "dry_run": self.dry_run,
                "repo_name": meta.get("repository"),
                "sca": meta.get("is_sca_scan", False),
                "full_scan": meta.get("is_full_scan", False),
                "semgrep_version": meta.get("semgrep_version", "0.0.0"),
            }
        )

        if self.dry_run:
            app_get_config_url = f"{state.env.semgrep_url}/{DEFAULT_SEMGREP_APP_CONFIG_URL}?{self._scan_params}"
        else:
            app_get_config_url = f"{state.env.semgrep_url}/{DEFAULT_SEMGREP_APP_CONFIG_URL}?{self._scan_params}"
            # TODO: uncomment the line below to replace the old endpoint with the new one once we have the
            # CLI logic in place to ignore findings that are from old rule versions
            # app_get_config_url = f"{state.env.semgrep_url}/api/agent/deployments/scans/{self.scan_id}/config"

        body = self._get_scan_config_from_app(app_get_config_url)

        self._deployment_id = body["deployment_id"]
        self._deployment_name = body["deployment_name"]
        self._policy_names = body["policy_names"]
        self._rules = body["rule_config"]
        self._autofix = body.get("autofix") or False
        self._deepsemgrep = body.get("deepsemgrep") or False
        self._dependency_query = body.get("dependency_query") or False
        self._skipped_syntactic_ids = body.get("triage_ignored_syntactic_ids") or []
        self._skipped_match_based_ids = body.get("triage_ignored_match_based_ids") or []
        self.ignore_patterns = body.get("ignored_files") or []

        if state.terminal.is_debug:
            config = deepcopy(body)
            try:
                config["rule_config"] = json.loads(config["rule_config"])
            except Exception:
                pass
            logger.debug(f"Got configuration {json.dumps(config, indent=4)}")

    def start_scan(self, meta: Dict[str, Any]) -> None:
        """
        Get scan id and file ignores

        returns ignored list
        """
        state = get_state()
        if self.dry_run:
            logger.info(f"Would have sent POST request to create scan")
            return

        logger.debug("Starting scan")
        response = state.app_session.post(
            f"{state.env.semgrep_url}/api/agent/deployments/scans",
            json={"meta": meta},
        )

        if response.status_code == 404:
            raise Exception(
                "Failed to create a scan with given token and deployment_id."
                "Please make sure they have been set correctly."
                f"API server at {state.env.semgrep_url} returned this response: {response.text}"
            )

        try:
            response.raise_for_status()
        except requests.RequestException:
            raise Exception(
                f"API server at {state.env.semgrep_url} returned this error: {response.text}"
            )

        body = response.json()
        self.scan_id = body["scan"]["id"]

    def report_failure(self, exit_code: int) -> None:
        """
        Send semgrep cli non-zero exit code information to server
        and return what exit code semgrep should exit with.
        """
        state = get_state()
        if self.dry_run:
            logger.info(f"Would have reported failure to semgrep.dev: {exit_code}")
            return

        response = state.app_session.post(
            f"{state.env.semgrep_url}/api/agent/scans/{self.scan_id}/error",
            json={
                "exit_code": exit_code,
                "stderr": "",
            },
        )

        try:
            response.raise_for_status()
        except requests.RequestException:
            raise Exception(f"API server returned this error: {response.text}")

    def report_findings(
        self,
        matches_by_rule: RuleMatchMap,
        errors: List[SemgrepError],
        rules: List[Rule],
        targets: Set[Path],
        renamed_targets: Set[Path],
        ignored_targets: FrozenSet[Path],
        parse_rate: ParsingData,
        total_time: float,
        commit_date: str,
        lockfile_dependencies: Dict[str, List[out.FoundDependency]],
        dependency_parser_errors: List[DependencyParserError],
        engine_requested: "EngineType",
    ) -> Tuple[bool, str]:
        """
        commit_date here for legacy reasons. epoch time of latest commit
        """
        state = get_state()
        rule_ids = [r.id for r in rules]
        all_matches = [
            match
            for matches_of_rule in matches_by_rule.values()
            for match in matches_of_rule
        ]
        # we want date stamps assigned by the app to be assigned such that the
        # current sort by relevant_since results in findings within a given scan
        # appear in an intuitive order.  this requires reversed ordering here.
        all_matches.reverse()
        sort_order = {  # used only to order rules by severity
            "EXPERIMENT": 0,
            "INVENTORY": 1,
            "INFO": 2,
            "WARNING": 3,
            "ERROR": 4,
        }
        # NB: sorted guarantees stable sort, so within a given severity level
        # issues remain sorted as before
        all_matches = sorted(
            all_matches, key=lambda match: sort_order[match.severity.value]
        )
        new_ignored, new_matches = partition(
            all_matches, lambda match: bool(match.is_ignored)
        )
        findings = [match.to_app_finding_format(commit_date) for match in new_matches]
        ignores = [
            match.to_app_finding_format(commit_date).to_json() for match in new_ignored
        ]
        token = (
            # GitHub (cloud)
            os.getenv("GITHUB_TOKEN")
            # GitLab.com (cloud)
            or os.getenv("GITLAB_TOKEN")
            # Bitbucket Cloud
            or os.getenv("BITBUCKET_TOKEN")
        )

        api_scans_findings = out.ApiScansFindings(
            # send a backup token in case the app is not available
            token=token,
            findings=findings,
            searched_paths=[str(t) for t in sorted(targets)],
            rule_ids=rule_ids,
            gitlab_token=None,
        )
        # TODO: add those fields in semgrep_output_v1.atd spec
        findings_and_ignores = {
            **api_scans_findings.to_json(),
            "renamed_paths": [str(rt) for rt in sorted(renamed_targets)],
            "ignores": ignores,
        }

        if any(match.severity == RuleSeverity.EXPERIMENT for match in new_ignored):
            logger.info("Some experimental rules were run during execution.")

        ignored_ext_freqs = Counter(
            [os.path.splitext(path)[1] for path in ignored_targets]
        )
        ignored_ext_freqs.pop("", None)  # don't count files with no extension

        dependency_counts = {k: len(v) for k, v in lockfile_dependencies.items()}

        complete = {
            "exit_code": 1
            if any(match.is_blocking and not match.is_ignored for match in all_matches)
            else 0,
            "dependency_parser_errors": [e.to_json() for e in dependency_parser_errors],
            "stats": {
                "findings": len(new_matches),
                "errors": [error.to_dict() for error in errors],
                "total_time": total_time,
                "unsupported_exts": dict(ignored_ext_freqs),
                "lockfile_scan_info": dependency_counts,
                "parse_rate": {
                    lang: {
                        "targets_parsed": data.num_targets - data.targets_with_errors,
                        "num_targets": data.num_targets,
                        "bytes_parsed": data.num_bytes - data.error_bytes,
                        "num_bytes": data.num_bytes,
                    }
                    for (lang, data) in parse_rate.get_errors_by_lang().items()
                },
                "engine_requested": engine_requested.name,
            },
        }

        if self._dependency_query:
            lockfile_dependencies_json = {}
            for path, dependencies in lockfile_dependencies.items():
                lockfile_dependencies_json[path] = [
                    dependency.to_json() for dependency in dependencies
                ]
            complete["dependencies"] = lockfile_dependencies_json

        if self.dry_run:
            logger.info(
                f"Would have sent findings and ignores blob: {json.dumps(findings_and_ignores, indent=4)}"
            )
            logger.info(
                f"Would have sent complete blob: {json.dumps(complete, indent=4)}"
            )
            return (False, "")
        else:
            logger.debug(
                f"Sending findings and ignores blob: {json.dumps(findings_and_ignores, indent=4)}"
            )
            logger.debug(f"Sending complete blob: {json.dumps(complete, indent=4)}")

        response = state.app_session.post(
            f"{state.env.semgrep_url}/api/agent/scans/{self.scan_id}/findings_and_ignores",
            timeout=state.env.upload_findings_timeout,
            json=findings_and_ignores,
        )

        try:
            response.raise_for_status()

            resp_errors = response.json()["errors"]
            for error in resp_errors:
                message = error["message"]
                click.echo(f"Server returned following warning: {message}", err=True)

        except requests.RequestException:
            raise Exception(f"API server returned this error: {response.text}")

        # mark as complete
        response = state.app_session.post(
            f"{state.env.semgrep_url}/api/agent/scans/{self.scan_id}/complete",
            timeout=state.env.upload_findings_timeout,
            json=complete,
        )

        try:
            response.raise_for_status()
        except requests.RequestException:
            raise Exception(
                f"API server at {state.env.semgrep_url} returned this error: {response.text}"
            )

        ret = response.json()
        return (ret.get("app_block_override", False), ret.get("app_block_reason", ""))



import json
import os
import subprocess
import sys
from pathlib import Path
from typing import NoReturn
from typing import Optional
from typing import Sequence

import click

from semgrep.commands.wrapper import handle_command_errors
from semgrep.error import SemgrepError
from semgrep.state import get_state
from semgrep.types import JsonObject
from semgrep.util import git_check_output


@click.command()
@click.option(
    "--message",
    "-m",
    prompt="Please provide a message describing the false negative",
    type=str,
    help="Explain what should have been found, plus any supporting information (such as framework, etc.).",
)
@click.option(
    "--email",
    type=str,
    help="Your email (defaults to your git email)",
)
@click.option("--start", "-s", type=int, help="Line at which vulnerability starts")
@click.option(
    "--end",
    "-e",
    type=int,
    help="Line at which vulnerability ends (defaults to --start)",
)
@click.option(
    "--yes", "-y", is_flag=True, help="Send data to Semgrep.dev without confirmation"
)
@click.argument("path", required=True, nargs=1, type=Path)
@handle_command_errors
def shouldafound(
    message: str,
    email: Optional[str],
    start: Optional[int],
    end: Optional[int],
    yes: bool,
    path: Path,
) -> NoReturn:
    """
    Report a false negative in this project. "path" should be the file in which you expected the vulnerability to be found.
    """
    env = get_state().env
    # try to set an email unless specifically asked not to.
    if not email and not env.shouldafound_no_email:
        try:
            email = _get_git_email()
        except subprocess.CalledProcessError:
            email = None

    text = _read_lines(path, start, end)

    data = {
        "email": email,
        "lines": text,
        "message": message,
        "path": str(path.resolve().relative_to(os.getcwd())),
    }

    if not yes:
        click.echo(
            "This information, including the following lines of code, will be sent to Semgrep.dev:",
            err=True,
        )
        click.echo(json.dumps(data, indent=2), err=True)
        if not click.confirm(
            "Please confirm the code does not contain any sensitive information, and you wish to send!",
            err=True,
        ):
            click.echo("Aborted", err=True)
            sys.exit(0)

    # send to backend
    try:
        playground_link = _make_shouldafound_request(data)
        click.echo("Sent feedback. Thanks for your contribution!")
        click.echo(
            f"You can view and extend the generated rule template here: {playground_link}"
        )
        sys.exit(0)
    except SemgrepError:
        click.echo(
            "Could not send feedback to server. Please consider instead reaching out to us another way!",
            err=True,
        )
        sys.exit(2)


def _get_git_email() -> str:
    """
    :raises CalledProcessError: If not in a git environment / email not configured
    """
    return git_check_output(["git", "config", "user.email"])


def _make_shouldafound_request(data: JsonObject) -> Optional[str]:
    state = get_state()
    resp = state.app_session.post(
        f"{state.env.shouldafound_base_url}/shouldafound", json=data
    )

    if resp.status_code == 200:
        if "playground_link" in resp.json():
            return str(resp.json()["playground_link"])
        else:
            raise SemgrepError(
                f"Failed to parse playground link, response text: {resp.text}"
            )
    else:
        raise SemgrepError(
            f"Failed to POST shouldafound data, error code: {resp.status_code}"
        )


def _read_lines(path: Path, start: Optional[int], end: Optional[int]) -> Sequence[str]:
    with path.open("r") as fd:
        lines = fd.readlines()
    if start is not None:
        if start < 1:
            click.echo("--start must be > 0", err=True)
            sys.exit(2)
        if start > len(lines):
            click.echo("--start must be no more than number of lines in file", err=True)
            sys.exit(2)
        if end is None:
            end = start
        else:
            if end < start:
                click.echo("--end must be >= than --start", err=True)
                sys.exit(2)
        lines = lines[start - 1 : end]
    text = "".join(lines)
    return text



import atexit
import os
import sys
import time
from collections import defaultdict
from pathlib import Path
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import click
from rich.padding import Padding
from rich.progress import Progress
from rich.progress import SpinnerColumn
from rich.progress import TextColumn
from rich.table import Table

import semgrep.semgrep_main
from semgrep.app import auth
from semgrep.app.scans import ScanHandler
from semgrep.commands.install import run_install_semgrep_pro
from semgrep.commands.scan import CONTEXT_SETTINGS
from semgrep.commands.scan import scan_options
from semgrep.commands.wrapper import handle_command_errors
from semgrep.console import console
from semgrep.console import Title
from semgrep.constants import OutputFormat
from semgrep.engine import EngineType
from semgrep.error import FATAL_EXIT_CODE
from semgrep.error import INVALID_API_KEY_EXIT_CODE
from semgrep.error import SemgrepError
from semgrep.ignores import IGNORE_FILE_NAME
from semgrep.meta import generate_meta_from_environment
from semgrep.meta import GithubMeta
from semgrep.meta import GitMeta
from semgrep.metrics import MetricsState
from semgrep.output import OutputHandler
from semgrep.output import OutputSettings
from semgrep.project import ProjectConfig
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatchMap
from semgrep.state import get_state
from semgrep.util import git_check_output
from semgrep.util import unit_str
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)

# These patterns are excluded via --exclude regardless of other ignore configuration
ALWAYS_EXCLUDE_PATTERNS = [".semgrep/", ".semgrep_logs/"]

# These patterns are excluded via --exclude unless the user provides their own .semgrepignore
DEFAULT_EXCLUDE_PATTERNS = ["test/", "tests/", "*_test.go"]


def yield_valid_patterns(patterns: Iterable[str]) -> Iterable[str]:
    """
    Parses patterns from semgrep.dev and returns the lines that
    are non-empty and do not start with #
    """
    for pattern in patterns:
        pattern = pattern.strip()

        if not pattern:
            continue
        if pattern.startswith("#"):
            continue

        yield pattern


def yield_exclude_paths(requested_patterns: Sequence[str]) -> Iterable[str]:
    patterns = [*yield_valid_patterns(requested_patterns), *ALWAYS_EXCLUDE_PATTERNS]
    if Path(IGNORE_FILE_NAME).is_file() and not requested_patterns:
        patterns.extend(DEFAULT_EXCLUDE_PATTERNS)

    yield from patterns


def fix_head_if_github_action(metadata: GitMeta) -> None:
    """
    GHA can checkout the incorrect commit for a PR (it will create a fake merge commit),
    so we need to reset the head to the actual PR branch head before continuing.

    Assumes cwd is a valid git project and that if we are in github-actions pull_request,
    metadata.head_branch_hash point to head commit of current branch
    """
    if not (isinstance(metadata, GithubMeta) and metadata.is_pull_request_event):
        return

    assert metadata.head_branch_hash is not None  # Not none when github action PR
    assert metadata.base_branch_hash is not None

    logger.info("Fixing git state for github action pull request")

    logger.debug("Calling git rev-parse HEAD")
    stashed_rev = git_check_output(["git", "rev-parse", "HEAD"])
    logger.debug(f"stashed_rev: {stashed_rev}")

    logger.info(f"Not on head ref: {metadata.head_branch_hash}; checking that out now.")
    git_check_output(["git", "checkout", metadata.head_branch_hash])

    atexit.register(git_check_output, ["git", "checkout", stashed_rev])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.pass_context
@scan_options
@click.option(
    "--audit-on",
    envvar="SEMGREP_AUDIT_ON",
    multiple=True,
    type=str,
    hidden=True,
)
@click.option(
    "--config",
    "-c",
    "-f",
    multiple=True,
    help="""
        YAML configuration file, directory of YAML files ending in
        .yml|.yaml, URL of a configuration file, or Semgrep registry entry name.
        \n\n
        Use --config auto to automatically obtain rules tailored to this project; your project URL will be used to log in
         to the Semgrep registry.
        \n\n
        To run multiple rule files simultaneously, use --config before every YAML, URL, or Semgrep registry entry name.
         For example `semgrep --config p/python --config myrules/myrule.yaml`
        \n\n
        See https://semgrep.dev/docs/writing-rules/rule-syntax for information on configuration file format.
    """,
    envvar="SEMGREP_RULES",
)
@click.option(
    "--dry-run",
    is_flag=True,
    help="""
        When set, will not start a scan on semgrep.dev and will not report findings.
        Instead will print out json objects it would have sent.
    """,
)
@click.option(
    "--supply-chain",
    is_flag=True,
    hidden=True,
)
@click.option(
    "--suppress-errors/--no-suppress-errors",
    "suppress_errors",
    default=True,
    help="""
        Configures how the CI command reacts when an error occurs.
        If true, encountered errors are suppressed and the exit code is zero (success).
        If false, encountered errors are not suppressed and the exit code is non-zero (success).
    """,
    envvar="SEMGREP_SUPPRESS_ERRORS",
)
@handle_command_errors
def ci(
    ctx: click.Context,
    *,
    audit_on: Sequence[str],
    autofix: bool,
    baseline_commit: Optional[str],
    core_opts: Optional[str],
    config: Optional[Tuple[str, ...]],
    debug: bool,
    dump_command_for_core: bool,
    dry_run: bool,
    enable_nosem: bool,
    enable_version_check: bool,
    exclude: Optional[Tuple[str, ...]],
    exclude_rule: Optional[Tuple[str, ...]],
    suppress_errors: bool,
    force_color: bool,
    include: Optional[Tuple[str, ...]],
    jobs: int,
    max_chars_per_line: int,
    max_lines_per_finding: int,
    max_memory: Optional[int],
    max_target_bytes: int,
    metrics: Optional[MetricsState],
    metrics_legacy: Optional[MetricsState],
    optimizations: str,
    dataflow_traces: Optional[bool],
    output: Optional[str],
    output_format: OutputFormat,
    requested_engine: EngineType,
    quiet: bool,
    rewrite_rule_ids: bool,
    supply_chain: bool,
    scan_unknown_extensions: bool,
    time_flag: bool,
    timeout_threshold: int,
    timeout: int,
    interfile_timeout: Optional[int],
    use_git_ignore: bool,
    verbose: bool,
) -> None:
    """
    The recommended way to run semgrep in CI

    In pull_request/merge_request (PR/MR) contexts, `semgrep ci` will only report findings
    that were introduced by the PR/MR.

    When logged in, `semgrep ci` runs rules configured on Semgrep App and sends findings
    to your findings dashboard.

    Only displays findings that were marked as blocking.
    """
    state = get_state()
    state.terminal.configure(
        verbose=verbose,
        debug=debug,
        quiet=quiet,
        force_color=force_color,
        output_format=output_format,
    )

    state.metrics.configure(metrics, metrics_legacy)
    state.error_handler.configure(suppress_errors)
    scan_handler = None

    token = state.app_session.token
    if not token and not config:
        # Not logged in and no explicit config
        logger.info("run `semgrep login` before using `semgrep ci` or set `--config`")
        sys.exit(INVALID_API_KEY_EXIT_CODE)
    elif not token and config:
        # Not logged in but has explicit config
        pass
    elif token and config:
        # Logged in but has explicit config
        logger.info(
            "Cannot run `semgrep ci` with --config while logged in. The `semgrep ci` command will upload findings to semgrep-app and those findings must come from rules configured there. Drop the `--config` to use rules configured on semgrep.dev or log out."
        )
        sys.exit(FATAL_EXIT_CODE)
    elif token:
        deployment_name = auth.get_deployment_from_token(token)
        if not deployment_name:
            logger.info(
                "API token not valid. Try to run `semgrep logout` and `semgrep login` again.",
            )
            sys.exit(INVALID_API_KEY_EXIT_CODE)
        scan_handler = ScanHandler(dry_run=dry_run, deployment_name=deployment_name)
    else:  # impossible state… until we break the code above
        raise RuntimeError("The token and/or config are misconfigured")

    output_settings = OutputSettings(
        output_format=output_format,
        output_destination=output,
        verbose_errors=verbose,
        timeout_threshold=timeout_threshold,
        output_time=time_flag,
        output_per_finding_max_lines_limit=max_lines_per_finding,
        output_per_line_max_chars_limit=max_chars_per_line,
    )
    output_handler = OutputHandler(output_settings)
    metadata = generate_meta_from_environment(baseline_commit)

    console.print(Title("Debugging Info"))
    debugging_table = Table.grid(padding=(0, 1))
    debugging_table.add_row(
        "versions",
        "-",
        f"semgrep [bold]{semgrep.__VERSION__}[/bold] on python [bold]{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}[/bold]",
    )
    debugging_table.add_row(
        "environment",
        "-",
        f"running in environment [bold]{metadata.environment}[/bold], triggering event is [bold]{metadata.event_name}[/bold]",
    )

    console.print(Title("Scan Environment", order=2))
    console.print(debugging_table, markup=True)

    fix_head_if_github_action(metadata)

    try:
        # Note this needs to happen within fix_head_if_github_action
        # so that metadata of current commit is correct
        if scan_handler:
            console.print(Title("Connection", order=2))
            metadata_dict = metadata.to_dict()
            metadata_dict["is_sca_scan"] = supply_chain
            proj_config = ProjectConfig.load_all()
            metadata_dict = {**metadata_dict, **proj_config.to_dict()}
            with Progress(
                TextColumn("  {task.description}"),
                SpinnerColumn(spinner_name="simpleDotsScrolling"),
                console=console,
            ) as progress_bar:
                at_url_maybe = (
                    f" at [bold]{state.env.semgrep_url}[/bold]"
                    if state.env.semgrep_url != "https://semgrep.dev"
                    else ""
                )

                start_scan_task = progress_bar.add_task(
                    f"Reporting start of scan for [bold]{scan_handler.deployment_name}[/bold]"
                )
                scan_handler.start_scan(metadata_dict)
                progress_bar.update(start_scan_task, completed=100)

                connection_task = progress_bar.add_task(
                    f"Fetching configuration from Semgrep Cloud Platform{at_url_maybe}"
                )
                scan_handler.fetch_and_init_scan_config(metadata_dict)
                progress_bar.update(connection_task, completed=100)

            config = (scan_handler.rules,)

    except Exception as e:
        import traceback

        traceback.print_exc()
        logger.info(f"Could not start scan {e}")
        sys.exit(FATAL_EXIT_CODE)

    engine_type = EngineType.decide_engine_type(
        requested_engine=requested_engine,
        scan_handler=scan_handler,
        git_meta=metadata,
    )

    # set default settings for selected engine type
    if dataflow_traces is None:
        dataflow_traces = engine_type.has_dataflow_traces

    if max_memory is None:
        max_memory = engine_type.default_max_memory

    if interfile_timeout is None:
        interfile_timeout = engine_type.default_interfile_timeout

    if engine_type.is_pro:
        console.print(Padding(Title("Engine", order=2), (1, 0, 0, 0)))
        if engine_type.check_if_installed():
            console.print(
                f"Using Semgrep Pro Version: [bold]{engine_type.get_pro_version()}[/bold]",
                markup=True,
            )
            console.print(
                f"Installed at [bold]{engine_type.get_binary_path()}[/bold]",
                markup=True,
            )
        else:
            run_install_semgrep_pro()

    try:
        excludes_from_app = scan_handler.ignore_patterns if scan_handler else []

        assert exclude is not None  # exclude is default empty tuple
        exclude = (*exclude, *yield_exclude_paths(excludes_from_app))
        assert config  # Config has to be defined here. Helping mypy out
        start = time.time()
        (
            filtered_matches_by_rule,
            semgrep_errors,
            renamed_targets,
            ignore_log,
            filtered_rules,
            profiler,
            output_extra,
            shown_severities,
            dependencies,
            dependency_parser_errors,
        ) = semgrep.semgrep_main.main(
            core_opts_str=core_opts,
            engine_type=engine_type,
            output_handler=output_handler,
            target=[os.curdir],  # semgrep ci only scans cwd
            pattern=None,
            lang=None,
            configs=config,
            no_rewrite_rule_ids=(not rewrite_rule_ids),
            dump_command_for_core=dump_command_for_core,
            jobs=jobs,
            include=include,
            exclude=exclude,
            exclude_rule=exclude_rule,
            max_target_bytes=max_target_bytes,
            autofix=scan_handler.autofix if scan_handler else False,
            dryrun=True,
            # Always true, as we want to always report all findings, even
            # ignored ones, to the backend
            disable_nosem=True,
            no_git_ignore=(not use_git_ignore),
            timeout=timeout,
            max_memory=max_memory,
            interfile_timeout=interfile_timeout,
            timeout_threshold=timeout_threshold,
            skip_unknown_extensions=(not scan_unknown_extensions),
            optimizations=optimizations,
            baseline_commit=metadata.merge_base_ref,
            baseline_commit_is_mergebase=True,
        )
    except SemgrepError as e:
        output_handler.handle_semgrep_errors([e])
        output_handler.output({}, all_targets=set(), filtered_rules=[])
        logger.info(f"Encountered error when running rules: {e}")
        if isinstance(e, SemgrepError):
            exit_code = e.code
        else:
            exit_code = FATAL_EXIT_CODE
        if scan_handler:
            scan_handler.report_failure(exit_code)
        sys.exit(exit_code)

    total_time = time.time() - start

    # Split up rules into respective categories:
    blocking_rules: List[Rule] = []
    nonblocking_rules: List[Rule] = []
    cai_rules: List[Rule] = []
    for rule in filtered_rules:
        if "r2c-internal-cai" in rule.id:
            cai_rules.append(rule)
        else:
            if rule.is_blocking:
                blocking_rules.append(rule)
            else:
                nonblocking_rules.append(rule)

    # Split up matches into respective categories
    blocking_matches_by_rule: RuleMatchMap = defaultdict(list)
    nonblocking_matches_by_rule: RuleMatchMap = defaultdict(list)
    cai_matches_by_rule: RuleMatchMap = defaultdict(list)

    # Since we keep nosemgrep disabled for the actual scan, we have to apply
    # that flag here
    keep_ignored = not enable_nosem or output_handler.formatter.keep_ignores()
    for rule, matches in filtered_matches_by_rule.items():
        # Filter out any matches that are triaged as ignored on the app
        if scan_handler:
            matches = [
                match
                for match in matches
                if match.syntactic_id not in scan_handler.skipped_syntactic_ids
                and match.match_based_id not in scan_handler.skipped_match_based_ids
            ]

        for match in matches:
            if match.is_ignored and not keep_ignored:
                continue

            applicable_result_set = (
                cai_matches_by_rule
                if "r2c-internal-cai" in rule.id
                else blocking_matches_by_rule
                # note that SCA findings are always non-blocking
                if rule.is_blocking and "sca_info" not in match.extra
                else nonblocking_matches_by_rule
            )
            applicable_result_set[rule].append(match)

    num_nonblocking_findings = sum(len(v) for v in nonblocking_matches_by_rule.values())
    num_blocking_findings = sum(len(v) for v in blocking_matches_by_rule.values())

    output_handler.output(
        {**blocking_matches_by_rule, **nonblocking_matches_by_rule},
        all_targets=output_extra.all_targets,
        ignore_log=ignore_log,
        profiler=profiler,
        filtered_rules=filtered_rules,
        profiling_data=output_extra.profiling_data,
        severities=shown_severities,
        is_ci_invocation=True,
        rules_by_engine=output_extra.rules_by_engine,
        engine_type=engine_type,
    )

    logger.info("CI scan completed successfully.")
    logger.info(
        f"  Found {unit_str(num_blocking_findings + num_nonblocking_findings, 'finding')} ({num_blocking_findings} blocking) from {unit_str(len(blocking_rules) + len(nonblocking_rules), 'rule')}."
    )

    app_block_override = False
    reason = ""
    if scan_handler:
        logger.info("  Uploading findings.")
        app_block_override, reason = scan_handler.report_findings(
            filtered_matches_by_rule,
            semgrep_errors,
            filtered_rules,
            output_extra.all_targets,
            renamed_targets,
            ignore_log.unsupported_lang_paths,
            output_extra.parsing_data,
            total_time,
            metadata.commit_datetime,
            dependencies,
            dependency_parser_errors,
            engine_type,
        )
        logger.info("  View results in Semgrep Cloud Platform:")
        logger.info(
            f"    https://semgrep.dev/orgs/{scan_handler.deployment_name}/findings"
        )
        if "r2c-internal-project-depends-on" in scan_handler.rules:
            logger.info(
                f"    https://semgrep.dev/orgs/{scan_handler.deployment_name}/supply-chain"
            )

    audit_mode = metadata.event_name in audit_on
    if num_blocking_findings > 0:
        if audit_mode:
            logger.info(
                f"  Audit mode is on for {metadata.event_name}, so exiting with code 0 even if matches found",
            )
            exit_code = 0
        else:
            logger.info("  Has findings for blocking rules so exiting with code 1")
            exit_code = 1
    else:
        logger.info("  No blocking findings so exiting with code 0")
        exit_code = 0

    if app_block_override and not audit_mode:
        logger.info(f"  semgrep.dev is suggesting a non-zero exit code ({reason})")
        exit_code = 1

    if enable_version_check:
        from semgrep.app.version import version_check

        version_check()

    sys.exit(exit_code)



import json
import urllib
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Union

from semgrep.app import auth
from semgrep.app.scans import ScanHandler
from semgrep.config_resolver import get_config
from semgrep.constants import DEFAULT_MAX_TARGET_SIZE
from semgrep.constants import DEFAULT_TIMEOUT
from semgrep.engine import EngineType
from semgrep.error import SemgrepError
from semgrep.meta import generate_meta_from_environment
from semgrep.meta import GitMeta
from semgrep.metrics import MetricsState
from semgrep.project import get_project_url
from semgrep.rule import Rule
from semgrep.semgrep_main import get_file_ignore
from semgrep.state import get_state
from semgrep.target_manager import TargetManager
from semgrep.types import JsonObject


class LSPConfig:
    def __init__(
        self, lsp_config: JsonObject, workspace_folders: List[JsonObject]
    ) -> None:
        lsp_config = dict(lsp_config)
        # Ensure that we have these config keys
        # This is easier than checking if they are None everywhere
        if "scan" not in lsp_config:
            lsp_config["scan"] = {}

        self._settings: Mapping[str, Mapping[str, Any]] = lsp_config
        self._workspace_folders: List[JsonObject] = workspace_folders
        self.rules: List[Rule] = []
        self.refresh_target_manager()

    # =====================
    # Semgrep scan settings
    # =====================

    @property
    def jobs(self) -> int:
        return self._settings["scan"].get("jobs", 1)

    @property
    def exclude(self) -> List[str]:
        return self._settings["scan"].get("exclude", [])

    @property
    def include(self) -> List[str]:
        return self._settings["scan"].get("include", [])

    @property
    def max_memory(self) -> int:
        return self._settings["scan"].get("maxMemory", 0)

    @property
    def max_target_bytes(self) -> int:
        return self._settings["scan"].get("maxTargetBytes", DEFAULT_MAX_TARGET_SIZE)

    @property
    def timeout(self) -> int:
        return self._settings["scan"].get("timeout", DEFAULT_TIMEOUT)

    @property
    def timeout_threshold(self) -> int:
        return self._settings["scan"].get("timeoutThreshold", 3)

    @property
    def project_url(self) -> Union[str, None]:
        return get_project_url()

    @property
    def ci_rules(self) -> List[Rule]:
        if len(self._workspace_folders) == 0:
            return []
        scan_handler = ScanHandler(True)
        metadata = generate_meta_from_environment(None)
        metadata_dict = metadata.to_dict()
        scan_handler.fetch_and_init_scan_config(metadata_dict)
        json_rules = json.loads(scan_handler.rules).get("rules", [])
        rules = [Rule.from_json(r) for r in json_rules]
        return rules

    @property
    def local_rules(self) -> List[Rule]:
        configs = []
        settings_configs = self._settings["scan"].get("configuration")

        if settings_configs is not None:
            configs.extend(settings_configs)
        if len(configs) == 0:
            return []

        configs_obj, _ = get_config(None, None, configs, project_url=self.project_url)
        rules = configs_obj.get_rules(True)
        return rules

    @property
    def auto_rules(self) -> List[Rule]:
        configs_obj, _ = get_config(None, None, ["auto"], project_url=self.project_url)
        rules = configs_obj.get_rules(True)
        return rules

    # =====================
    # Semgrep LSP settings
    # =====================
    @property
    def extension_metrics(self) -> Dict[str, Any]:
        extension_metrics = self._settings.get("extensionMetrics", {})
        if isinstance(extension_metrics, dict):
            return extension_metrics
        elif isinstance(extension_metrics, str):
            # for backwards compatibility, older versions of the extension may send metrics: "on" or "off"
            return {
                "enabled": extension_metrics == "on",
            }
        else:
            return {}

    @property
    def metrics_state(self) -> MetricsState:
        choice = self.extension_metrics.get("enabled", True)
        if choice:
            return MetricsState.ON
        else:
            return MetricsState.OFF

    # =====================
    # Useful properties
    # =====================

    @property
    def settings(self) -> JsonObject:
        return self._settings

    @property
    def head_commit(self) -> Optional[str]:
        return GitMeta().commit_sha

    @property
    def folders(self) -> List[str]:
        """All workspace folders by URI"""
        return [f["uri"] for f in self._workspace_folders]

    @property
    def folder_paths(self) -> List[str]:
        """All workspace folders by path"""
        folder_paths = []
        for f in self.folders:
            uri = urllib.parse.urlparse(f)
            target_name = urllib.request.url2pathname(uri.path)
            folder_paths.append(target_name)
        return folder_paths

    @property
    def token(self) -> Optional[str]:
        return auth.get_token()

    @property
    def logged_in(self) -> bool:
        return (
            self.token is not None
            and auth.get_deployment_from_token(self.token) is not None
        )

    @property
    def debug(self) -> bool:
        trace = self._settings.get("trace")
        debug = trace.get("server") if trace else None
        return debug is not None and debug == "verbose"

    @property
    def engine_type(self) -> EngineType:
        # Pro engine is too slow right now for editor integration
        return EngineType.OSS

    # =====================
    # Config management
    # =====================

    def refresh_target_manager(self) -> None:
        self.target_manager = TargetManager(
            includes=self.include,
            excludes=self.exclude,
            max_target_bytes=self.max_target_bytes,
            respect_git_ignore=True,
            file_ignore=get_file_ignore(),
            target_strings=self.folder_paths,
        )

    def refresh_rules(self) -> None:
        self.rules = []
        logged_in = self.logged_in

        self.rules.extend(self.local_rules)
        if logged_in:
            self.rules.extend(self.ci_rules)
        elif len(self.rules) == 0:
            self.rules.extend(self.auto_rules)

    def send_metrics(
        self, exit_code: Optional[int] = None, error: Optional[SemgrepError] = None
    ) -> None:
        state = get_state()
        metrics = state.metrics
        metrics.configure(self.metrics_state, None)
        metrics.add_engine_type(self.engine_type)
        metrics.add_project_url(self.project_url)
        metrics.add_token(self.token)
        if exit_code is not None:
            metrics.add_exit_code(exit_code)
        if error is not None:
            metrics.add_errors([error])
        extension_metrics = self.extension_metrics
        machine_id = extension_metrics.get("machineId")
        new_install = extension_metrics.get("isNewAppInstall")
        session_id = extension_metrics.get("sessionId")
        version = extension_metrics.get("extensionVersion")
        type = extension_metrics.get("extensionType")
        metrics.add_extension(machine_id, new_install, session_id, version, type)
        metrics.send()



import textwrap
from contextlib import contextmanager
from itertools import groupby
from pathlib import Path
from shutil import get_terminal_size
from typing import Any
from typing import cast
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple

import click
import colorama
from rich.console import Console

import semgrep.semgrep_interfaces.semgrep_output_v1 as out
from semgrep.console import console
from semgrep.console import Title
from semgrep.constants import CLI_RULE_ID
from semgrep.constants import Colors
from semgrep.constants import ELLIPSIS_STRING
from semgrep.constants import MAX_CHARS_FLAG_NAME
from semgrep.constants import MAX_LINES_FLAG_NAME
from semgrep.error import SemgrepCoreError
from semgrep.error import SemgrepError
from semgrep.formatter.base import BaseFormatter
from semgrep.rule import Rule
from semgrep.rule import RuleProduct
from semgrep.rule_match import RuleMatch
from semgrep.semgrep_types import LANGUAGE
from semgrep.semgrep_types import Language
from semgrep.util import format_bytes
from semgrep.util import get_lines
from semgrep.util import truncate
from semgrep.util import unit_str
from semgrep.util import with_color

MAX_TEXT_WIDTH = 120

BASE_INDENT = 8

terminal_size = get_terminal_size((MAX_TEXT_WIDTH, 1))[0]
if terminal_size <= 0:
    terminal_size = MAX_TEXT_WIDTH
width = min(MAX_TEXT_WIDTH, terminal_size)
if width <= 110:
    width = width - 5
else:
    width = width - (width - 100)

FINDINGS_INDENT_DEPTH = 10


GROUP_TITLES: Dict[Tuple[RuleProduct, str], str] = {
    (RuleProduct.sca, "unreachable"): "Unreachable Supply Chain Finding",
    (RuleProduct.sca, "undetermined"): "Undetermined Supply Chain Finding",
    (RuleProduct.sca, "reachable"): "Reachable Supply Chain Finding",
    (RuleProduct.sast, "nonblocking"): "Non-blocking Code Finding",
    (RuleProduct.sast, "blocking"): "Blocking Code Finding",
    (RuleProduct.sast, "merged"): "Code Finding",
}


def color_line(
    line: str,
    line_number: int,
    start_line: int,
    start_col: int,
    end_line: int,
    end_col: int,
) -> str:
    """
    Assumes column start and end numbers are 1-indexed
    """
    start_color = 0 if line_number > start_line else start_col
    # adjust for 1-indexed column number
    start_color = max(start_color - 1, 0)
    # put the end color at the end of the line if this isn't the last line in the output
    end_color = end_col if line_number >= end_line else len(line) + 1 + 1
    # adjust for 1-indexed column number
    end_color = max(end_color - 1, 0)
    line = (
        line[:start_color]
        + with_color(Colors.foreground, line[start_color:end_color], bold=True)
        + line[end_color:]
    )
    return line


def format_lines(
    path: Path,
    start_line: int,
    start_col: int,
    end_line: int,
    end_col: int,
    lines: List[str],
    color_output: bool,
    per_finding_max_lines_limit: Optional[int],
    per_line_max_chars_limit: Optional[int],
    show_separator: bool,
    show_path: bool,
) -> Iterator[str]:
    trimmed = 0
    stripped = False

    if per_finding_max_lines_limit:
        trimmed = len(lines) - per_finding_max_lines_limit
        lines = lines[:per_finding_max_lines_limit]

    # we remove indentation at the start of the snippet to avoid wasting space
    dedented_lines = textwrap.dedent("".join(lines)).splitlines()
    indent_len = (
        len(lines[0].rstrip()) - len(dedented_lines[0].rstrip())
        if len(dedented_lines) > 0 and len(lines) > 0
        else 0
    )

    # since we dedented each line, we need to adjust where the highlighting is
    start_col -= indent_len
    end_col -= indent_len

    for i, line in enumerate(dedented_lines):
        line = line.rstrip()
        line_number = ""
        if start_line:
            if color_output:
                line = color_line(
                    line,
                    start_line + i,
                    start_line,
                    start_col,
                    end_line,
                    end_col,
                )
                line_number = f"{start_line + i}"
            else:
                line_number = f"{start_line + i}"

            if per_line_max_chars_limit and len(line) > per_line_max_chars_limit:
                stripped = True
                is_first_line = i == 0
                if is_first_line:
                    line = (
                        line[start_col - 1 : start_col - 1 + per_line_max_chars_limit]
                        + ELLIPSIS_STRING
                    )
                    if start_col > 1:
                        line = ELLIPSIS_STRING + line
                else:
                    line = line[:per_line_max_chars_limit] + ELLIPSIS_STRING
                # while stripping a string, the ANSI code for resetting color might also get stripped.
                line = line + colorama.Style.RESET_ALL

        # plus one because we want this to be slightly separated from the intervening messages
        if i == 0 and show_path:
            yield f" " * (
                BASE_INDENT + 1
            ) + f"{with_color(Colors.cyan, f'{path}', bold=False)}"

        yield f" " * (
            11 - len(line_number)
        ) + f"{line_number}┆ {line}" if line_number else f"{line}"

    if stripped:
        stripped_str = (
            f"[shortened a long line from output, adjust with {MAX_CHARS_FLAG_NAME}]"
        )
        yield " " * FINDINGS_INDENT_DEPTH + stripped_str

    if per_finding_max_lines_limit != 1:
        if trimmed > 0:
            trimmed_str = (
                f" [hid {trimmed} additional lines, adjust with {MAX_LINES_FLAG_NAME}] "
            )
            yield " " * FINDINGS_INDENT_DEPTH + trimmed_str
        elif lines and show_separator:
            yield f" " * FINDINGS_INDENT_DEPTH + f"⋮┆" + f"-" * 40


def finding_to_line(
    rule_match: RuleMatch,
    color_output: bool,
    per_finding_max_lines_limit: Optional[int],
    per_line_max_chars_limit: Optional[int],
    show_separator: bool,
) -> Iterator[str]:
    path = rule_match.path
    start_line = rule_match.start.line
    end_line = rule_match.end.line
    start_col = rule_match.start.col
    end_col = rule_match.end.col
    if path:
        lines = rule_match.extra.get("fixed_lines") or rule_match.lines
        yield from format_lines(
            path,
            start_line,
            start_col,
            end_line,
            end_col,
            lines,
            color_output,
            per_finding_max_lines_limit,
            per_line_max_chars_limit,
            show_separator,
            False,
        )


def match_to_lines(
    ref_path: Path,
    location: out.Location,
    content: str,
    color_output: bool,
    per_finding_max_lines_limit: Optional[int],
    per_line_max_chars_limit: Optional[int],
) -> Iterator[str]:
    path = Path(location.path.value)
    is_same_file = path == ref_path
    lines = get_lines(path, location.start.line, location.end.line)
    yield from format_lines(
        path,
        location.start.line,
        location.start.col,
        location.end.line,
        location.end.col,
        lines,
        color_output,
        per_finding_max_lines_limit,
        per_line_max_chars_limit,
        False,
        not is_same_file,
    )


def call_trace_to_lines(
    ref_path: Path,
    call_trace: out.CliMatchCallTrace,
    color_output: bool,
    per_finding_max_lines_limit: Optional[int],
    per_line_max_chars_limit: Optional[int],
) -> Iterator[str]:
    trace = call_trace.value
    if isinstance(trace, out.CliLoc):
        yield from match_to_lines(
            ref_path,
            trace.value[0],
            trace.value[1],
            color_output,
            per_finding_max_lines_limit,
            per_line_max_chars_limit,
        )

    elif isinstance(trace, out.CliCall):
        data, intermediate_vars, call_trace = trace.value

        yield from match_to_lines(
            ref_path,
            data[0],
            data[1],
            color_output,
            per_finding_max_lines_limit,
            per_line_max_chars_limit,
        )

        if intermediate_vars and len(intermediate_vars) > 0:
            # TODO change this message based on rule kind if we ever use
            # dataflow traces for more than just taint
            yield (
                BASE_INDENT * " " + "Taint flows through these intermediate variables:"
            )
            prev_path = ref_path
            for var in intermediate_vars:
                loc = var.location
                path = Path(loc.path.value)
                lines = get_lines(Path(loc.path.value), loc.start.line, loc.end.line)
                is_same_file = path == prev_path
                yield from format_lines(
                    Path(loc.path.value),
                    loc.start.line,
                    loc.start.col,
                    loc.end.line,
                    loc.end.col,
                    lines,
                    color_output,
                    per_finding_max_lines_limit,
                    per_line_max_chars_limit,
                    False,
                    not is_same_file,
                )
                prev_path = path

        if isinstance(call_trace.value, out.CliCall):
            yield (BASE_INDENT * " " + "then call to:")
        elif isinstance(call_trace.value, out.CliLoc):
            yield (BASE_INDENT * " " + "then reaches:")
        yield from call_trace_to_lines(
            ref_path,
            call_trace,
            color_output,
            per_finding_max_lines_limit,
            per_line_max_chars_limit,
        )


def dataflow_trace_to_lines(
    rule_match_path: Path,
    dataflow_trace: Optional[out.CliMatchDataflowTrace],
    color_output: bool,
    per_finding_max_lines_limit: Optional[int],
    per_line_max_chars_limit: Optional[int],
    show_separator: bool,
) -> Iterator[str]:
    if dataflow_trace:
        source = dataflow_trace.taint_source
        intermediate_vars = dataflow_trace.intermediate_vars
        sink = dataflow_trace.taint_sink

        if source:
            yield ""
            yield (BASE_INDENT * " " + "Taint comes from:")
            yield from call_trace_to_lines(
                rule_match_path,
                source,
                color_output,
                per_finding_max_lines_limit,
                per_line_max_chars_limit,
            )

        if intermediate_vars and len(intermediate_vars) > 0:
            # TODO change this message based on rule kind of we ever use
            # dataflow traces for more than just taint
            yield ""
            yield (
                BASE_INDENT * " " + "Taint flows through these intermediate variables:"
            )
            prev_path = rule_match_path
            for var in intermediate_vars:
                loc = var.location
                path = Path(loc.path.value)
                lines = get_lines(path, loc.start.line, loc.end.line)
                is_same_file = path == prev_path
                yield from format_lines(
                    path,
                    loc.start.line,
                    loc.start.col,
                    loc.end.line,
                    loc.end.col,
                    lines,
                    color_output,
                    per_finding_max_lines_limit,
                    per_line_max_chars_limit,
                    False,
                    not is_same_file,
                )
                prev_path = path

        if sink:
            yield ""
            yield (BASE_INDENT * " " + "This is how taint reaches the sink:")
            yield from call_trace_to_lines(
                rule_match_path,
                sink,
                color_output,
                per_finding_max_lines_limit,
                per_line_max_chars_limit,
            )
            yield ""

        if source and show_separator:
            yield f" " * BASE_INDENT + f"⋮┆" + f"-" * 40


def get_details_shortlink(rule_match: RuleMatch) -> Optional[str]:
    source_url = rule_match.metadata.get("shortlink")
    if not source_url:
        return ""
    return f"Details: {source_url}"


def print_time_summary(
    time_data: out.CliTiming, error_output: Sequence[SemgrepError]
) -> None:
    items_to_show = 5
    col_lim = 50

    targets = time_data.targets

    # Compute summary timings
    rule_parsing_time = time_data.rules_parse_time
    rule_match_timings = {
        rule.id.value: sum(t.match_times[i] for t in targets if t.match_times[i] >= 0)
        for i, rule in enumerate(time_data.rules)
    }
    file_parsing_time = sum(
        sum(t for t in target.parse_times if t >= 0) for target in targets
    )
    file_timings = {
        target.path.value: (
            sum(t for t in target.parse_times if t >= 0),
            target.run_time,
        )
        for target in targets
    }

    all_total_time = sum(i[1] for i in file_timings.values()) + rule_parsing_time
    total_matching_time = sum(i for i in rule_match_timings.values())

    # Count errors

    semgrep_core_errors = [
        cast(SemgrepCoreError, err)
        for err in error_output
        if SemgrepError.semgrep_error_type(err) == "SemgrepCoreError"
    ]
    errors = {
        (err.core.location.path.value, err.core.error_type.kind)
        for err in semgrep_core_errors
    }

    error_types = {k: len(list(v)) for k, v in groupby(errors, lambda x: x[1])}
    num_errors = len(errors)

    # Compute summary by language

    # TODO assumes languages correspond solely to extension
    # Consider: get a report from semgrep-core on what language each
    #           file was analyzed as
    # However, this might make it harder for users to confirm
    # semgrep counts against their expected file counts

    ext_to_lang: Mapping[str, Language] = LANGUAGE.lang_by_ext

    def lang_of_path(path: str) -> str:
        ext = "." + path.split(".")[-1]
        return ext_to_lang.get(ext, "generic")

    ext_info = sorted(
        (
            (
                lang_of_path(target.path.value),
                (target.num_bytes, target.run_time),
            )
            for target in targets
        ),
        key=lambda x: x[0],
    )
    lang_info = {k: list(v) for k, v in groupby(ext_info, lambda x: x[0])}
    langs = lang_info.keys()
    lang_counts: Mapping[str, int] = {lang: len(lang_info[lang]) for lang in langs}
    lang_bytes: Mapping[str, int] = {
        lang: sum(info[1][0] for info in lang_info[lang]) for lang in langs
    }
    lang_times: Mapping[str, float] = {
        lang: sum(info[1][1] for info in lang_info[lang]) for lang in langs
    }

    # Output semgrep summary
    total_time = time_data.profiling_times.get("total_time", 0.0)
    config_time = time_data.profiling_times.get("config_time", 0.0)
    core_time = time_data.profiling_times.get("core_time", 0.0)
    # time_data.profiling_times.get("ignores_time", 0.0)

    console.print(
        "\n============================[ summary ]============================"
    )

    console.print(
        f"Total time: {total_time:.4f}s Config time: {config_time:.4f}s Core time: {core_time:.4f}s"
    )

    # Output semgrep-core information
    console.print("\nSemgrep-core time:")
    console.print(
        f"Total CPU time: {all_total_time:.4f}s  File parse time: {file_parsing_time:.4f}s"
        f"  Rule parse time: {rule_parsing_time:.4f}s  Match time: {total_matching_time:.4f}s"
    )

    console.print(f"Slowest {items_to_show}/{len(file_timings)} files")
    slowest_file_times = sorted(file_timings.items(), key=lambda x: x[1], reverse=True)[
        :items_to_show
    ]
    for file_name, (parse_time, run_time) in slowest_file_times:
        num_bytes = f"({format_bytes(Path(file_name).resolve().stat().st_size)}):"
        file_name = truncate(file_name, col_lim)
        console.print(
            f"{with_color(Colors.green, f'{file_name:<50}')} {num_bytes:<8} {run_time:.3f}s ({parse_time:.3f}s to parse)"
        )

    console.print(f"Slowest {items_to_show} rules to match")
    slowest_rule_times = sorted(rule_match_timings.items(), reverse=True)[
        :items_to_show
    ]
    for rule_id, match_time in slowest_rule_times:
        rule_id = truncate(rule_id, col_lim) + ":"
        console.print(
            f"{with_color(Colors.yellow, f'{rule_id:<59}')} {match_time:.3f}s"
        )

    # Output other file information
    ANALYZED = "Analyzed:"
    FAILED = "Errors:"
    headings = [ANALYZED, FAILED]
    max_heading_len = max(len(h) for h in headings) + 1  # for the space

    def add_heading(heading: str, lines: List[str]) -> List[str]:
        heading = heading + " " * (max_heading_len - len(heading))
        first = True
        returned = []
        for line in lines:
            prefix = heading if first else " " * max_heading_len
            returned.append(f"{prefix}{line}")
            first = False
        return returned

    console.print()

    by_lang = [
        f"{ lang_counts[lang] } { lang } files ({ format_bytes(lang_bytes[lang]) } in {(lang_times[lang]):.3f} seconds)"
        for lang in langs
    ]
    console.print("\n".join(add_heading(ANALYZED, by_lang)))

    # Output errors
    def if_exists(num_errors: int, msg: str) -> str:
        return "" if num_errors == 0 else msg

    see_more = if_exists(
        num_errors,
        ", see output before the results for details or run with --strict",
    )
    error_msg = f"{ num_errors } files with errors{see_more}"
    error_lines = [error_msg] + [
        f"{type} ({num} files)" for (type, num) in error_types.items()
    ]
    console.print("\n".join(add_heading(FAILED, error_lines)))
    console.print()


def print_text_output(
    rule_matches: Iterable[RuleMatch],
    color_output: bool,
    per_finding_max_lines_limit: Optional[int],
    per_line_max_chars_limit: Optional[int],
    dataflow_traces: bool,
) -> None:
    last_file = None
    last_message = None
    sorted_rule_matches = sorted(rule_matches, key=lambda r: (r.path, r.rule_id))
    for rule_index, rule_match in enumerate(sorted_rule_matches):
        current_file = rule_match.path
        message = rule_match.message
        fix = rule_match.fix
        if "sca_info" in rule_match.extra and (rule_match.extra["sca_info"].reachable):
            lockfile = rule_match.extra["sca_info"].dependency_match.lockfile
        else:
            lockfile = None
        if last_file is None or last_file != current_file:
            if last_file is not None:
                console.print()
            console.print(
                f"\n{with_color(Colors.cyan, f'  {current_file} ', bold=False)}"
                + (
                    f"with lockfile {with_color(Colors.cyan, f'{lockfile}')}"
                    if lockfile
                    else ""
                )
            )
            last_message = None
        # don't display the rule line if the check is empty
        if (
            rule_match.rule_id
            and rule_match.rule_id != CLI_RULE_ID
            and (last_message is None or last_message != message)
        ):
            shortlink = get_details_shortlink(rule_match)
            shortlink_text = (8 * " " + shortlink + "\n") if shortlink else ""
            title_text = click.wrap_text(
                f"{with_color(Colors.foreground, rule_match.title, bold=True)}",
                width + 10,
                5 * " ",
                5 * " ",
                False,
            )
            message_text = click.wrap_text(f"{message}", width, 8 * " ", 8 * " ", True)
            console.print(f"{title_text}\n{message_text}\n{shortlink_text}")

        last_file = current_file
        last_message = message
        next_rule_match = (
            sorted_rule_matches[rule_index + 1]
            if rule_index != len(sorted_rule_matches) - 1
            else None
        )
        autofix_tag = with_color(Colors.green, "         ▶▶┆ Autofix ▶")
        if fix is not None:
            console.print(
                f"{autofix_tag} {fix if fix else with_color(Colors.red, 'delete')}"
            )
        elif rule_match.fix_regex:
            fix_regex = rule_match.fix_regex
            console.print(
                f"{autofix_tag} s/{fix_regex.regex}/{fix_regex.replacement}/{fix_regex.count or 'g'}"
            )

        is_same_file = (
            next_rule_match.path == rule_match.path if next_rule_match else False
        )
        for line in finding_to_line(
            rule_match,
            color_output,
            per_finding_max_lines_limit,
            per_line_max_chars_limit,
            # if we have dataflow traces on, then we should print the separator,
            # because otherwise it is easy to mistake taint traces as belonging
            # to a different finding
            is_same_file and not (dataflow_traces and rule_match.dataflow_trace),
        ):
            console.print(line)

        if dataflow_traces:
            for line in dataflow_trace_to_lines(
                rule_match.path,
                rule_match.dataflow_trace,
                color_output,
                per_finding_max_lines_limit,
                per_line_max_chars_limit,
                is_same_file,
            ):
                console.print("  " + line)


@contextmanager
def force_quiet_off(console: Console) -> Iterator[None]:
    """
    Force the console to be not quiet, even if it was set to be quiet before.
    """
    was_quiet = console.quiet
    console.quiet = False
    try:
        yield
    finally:
        console.quiet = was_quiet


class TextFormatter(BaseFormatter):
    def format(
        self,
        rules: Iterable[Rule],
        rule_matches: Iterable[RuleMatch],
        semgrep_structured_errors: Sequence[SemgrepError],
        cli_output_extra: out.CliOutputExtra,
        extra: Mapping[str, Any],
        is_ci_invocation: bool,
    ) -> str:
        # all output in this function is captured and returned as a string
        with force_quiet_off(console), console.capture() as captured_output:
            grouped_matches: Dict[Tuple[RuleProduct, str], List[RuleMatch]] = {
                # ordered most important to least important
                (RuleProduct.sast, "blocking"): [],
                (RuleProduct.sca, "reachable"): [],
                (RuleProduct.sca, "undetermined"): [],
                (RuleProduct.sca, "unreachable"): [],
                (RuleProduct.sast, "nonblocking"): [],
            }

            for match in rule_matches:
                if match.product == RuleProduct.sast:
                    subgroup = "blocking" if match.is_blocking else "nonblocking"
                else:
                    subgroup = match.exposure_type or "undetermined"

                grouped_matches[match.product, subgroup].append(match)

            first_party_blocking_rules = {
                match.match.rule_id.value
                for match in grouped_matches[RuleProduct.sast, "blocking"]
            }

            # When ephemeral rules are run with the -e or --pattern flag in the command-line, the rule_id is set to -.
            # The rule is ran in the command-line and has no associated rule_id
            first_party_blocking_rules.discard("-")

            if not is_ci_invocation:
                grouped_matches[(RuleProduct.sast, "merged")] = [
                    *grouped_matches.pop((RuleProduct.sast, "nonblocking")),
                    *grouped_matches.pop((RuleProduct.sast, "blocking")),
                ]

            for group, matches in grouped_matches.items():
                if not matches:
                    continue
                console.print(Title(unit_str(len(matches), GROUP_TITLES[group])))
                print_text_output(
                    matches,
                    extra.get("color_output", False),
                    extra["per_finding_max_lines_limit"],
                    extra["per_line_max_chars_limit"],
                    extra["dataflow_traces"],
                )

            if first_party_blocking_rules and is_ci_invocation:
                console.print(Title("Blocking Code Rules Fired:", order=2))
                for rule_id in sorted(first_party_blocking_rules):
                    console.print(f"  {rule_id}")
                console.reset_title(order=1)

            if cli_output_extra.time:
                print_time_summary(cli_output_extra.time, semgrep_structured_errors)

            rules_by_engine = (
                cli_output_extra.rules_by_engine
                if cli_output_extra.rules_by_engine
                else []
            )

            rules_ran_within_a_file = [
                with_color(Colors.foreground, rule.value[0].value, bold=True)
                for rule in rules_by_engine
                if isinstance(rule.value[1].value, out.OSS)
            ]

            if (extra["engine_requested"].is_interfile) and rules_ran_within_a_file:
                console.print(
                    f"{unit_str(len(rules_ran_within_a_file), 'rule')} ran in a within-a-file fashion"
                    + " because `interfile: true` was not specified."
                )
                if extra.get("verbose_errors"):
                    console.print(
                        "These rules were:\n   "
                        + "   \n   ".join(rules_ran_within_a_file)
                    )
                else:
                    console.print("(Use --verbose to see which ones.)")

        return captured_output.get()



from functools import lru_cache
from pathlib import Path
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Tuple

import semgrep.output_from_core as core
from semdep.external.packaging.specifiers import InvalidSpecifier  # type: ignore
from semdep.external.packaging.specifiers import SpecifierSet  # type: ignore
from semdep.package_restrictions import dependencies_range_match_any
from semdep.parse_lockfile import parse_lockfile_path
from semgrep.error import SemgrepError
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch
from semgrep.semgrep_interfaces.semgrep_output_v1 import DependencyMatch
from semgrep.semgrep_interfaces.semgrep_output_v1 import DependencyPattern
from semgrep.semgrep_interfaces.semgrep_output_v1 import Direct
from semgrep.semgrep_interfaces.semgrep_output_v1 import Ecosystem
from semgrep.semgrep_interfaces.semgrep_output_v1 import FoundDependency
from semgrep.semgrep_interfaces.semgrep_output_v1 import ScaInfo
from semgrep.semgrep_interfaces.semgrep_output_v1 import Transitive
from semgrep.semgrep_interfaces.semgrep_output_v1 import Transitivity
from semgrep.target_manager import TargetManager

SCA_FINDING_SCHEMA = 20220913


def parse_depends_on_yaml(entries: List[Dict[str, str]]) -> Iterator[DependencyPattern]:
    """
    Convert the entries in the Yaml to ProjectDependsOnEntry objects that specify
    namespace, package name, and semver ranges
    """
    for entry in entries:
        # schema checks should gaurantee we have these fields, but we'll code defensively
        namespace = entry.get("namespace")
        if namespace is None:
            raise SemgrepError(f"project-depends-on is missing `namespace`")
        try:
            ecosystem = Ecosystem.from_json(namespace.lower())
        except ValueError:
            raise SemgrepError(f"unknown package ecosystem: {namespace}")
        package = entry.get("package")
        if package is None:
            raise SemgrepError(f"project-depends-on is missing `package`")
        semver_range = entry.get("version")
        if semver_range is None:
            raise SemgrepError(f"project-depends-on is missing `version`")
        try:
            SpecifierSet(semver_range)
        except InvalidSpecifier:
            raise SemgrepError(f"invalid semver range {semver_range}")

        yield DependencyPattern(
            ecosystem=ecosystem, package=package, semver_range=semver_range
        )


def generate_unreachable_sca_findings(
    rule: Rule,
    target_manager: TargetManager,
    already_reachable: Callable[[Path, FoundDependency], bool],
) -> Tuple[List[RuleMatch], List[SemgrepError]]:
    """
    Returns matches to a only a rule's sca-depends-on patterns; ignoring any reachabiliy patterns it has
    """
    depends_on_keys = rule.project_depends_on
    dep_rule_errors: List[SemgrepError] = []

    depends_on_entries = list(parse_depends_on_yaml(depends_on_keys))
    ecosystems = list(rule.ecosystems)

    non_reachable_matches = []
    for ecosystem in ecosystems:
        lockfile_paths = target_manager.get_lockfiles(ecosystem)

        for lockfile_path in lockfile_paths:
            # Ignore errors here because we assume they are processed later
            deps, _ = parse_lockfile_path(lockfile_path)
            dependency_matches = list(
                dependencies_range_match_any(depends_on_entries, list(deps))
            )
            for dep_pat, found_dep in dependency_matches:
                if already_reachable(lockfile_path, found_dep):
                    continue
                dep_match = DependencyMatch(
                    dependency_pattern=dep_pat,
                    found_dependency=found_dep,
                    lockfile=str(lockfile_path),
                )
                match = RuleMatch(
                    message=rule.message,
                    metadata=rule.metadata,
                    severity=rule.severity,
                    fix=None,
                    fix_regex=None,
                    match=core.CoreMatch(
                        rule_id=core.RuleId(rule.id),
                        location=core.Location(
                            path=core.Fpath(str(lockfile_path)),
                            start=core.Position(found_dep.line_number or 0, 0, 0),
                            end=core.Position(
                                (
                                    found_dep.line_number + 1
                                    if found_dep.line_number
                                    else 0
                                ),
                                0,
                                0,
                            ),
                        ),
                        # TODO: we need to define the fields below in
                        # Output_from_core.atd so we can reuse core.MatchExtra
                        extra=core.CoreMatchExtra(
                            metavars=core.Metavars({}),
                            engine_kind=core.EngineKind(core.OSS()),
                        ),
                    ),
                    extra={
                        "sca_info": ScaInfo(
                            sca_finding_schema=SCA_FINDING_SCHEMA,
                            reachable=False,
                            reachability_rule=rule.should_run_on_semgrep_core,
                            dependency_match=dep_match,
                        )
                    },
                )
                non_reachable_matches.append(match)
    return non_reachable_matches, dep_rule_errors


@lru_cache(maxsize=100_000)
def transivite_dep_is_also_direct(
    package: str, deps: Tuple[Tuple[str, Transitivity], ...]
) -> bool:
    """
    Assumes that [dep] is transitive
    Checks if there is a direct version of the transitive dependency [dep]
    """
    return (package, Transitivity(Direct())) in deps


def generate_reachable_sca_findings(
    matches: List[RuleMatch], rule: Rule, target_manager: TargetManager
) -> Tuple[
    List[RuleMatch], List[SemgrepError], Callable[[Path, FoundDependency], bool]
]:
    depends_on_keys = rule.project_depends_on
    dep_rule_errors: List[SemgrepError] = []

    depends_on_entries = list(parse_depends_on_yaml(depends_on_keys))
    ecosystems = list(rule.ecosystems)

    # Reachability rule
    reachable_matches = []
    reachable_deps = set()
    for ecosystem in ecosystems:
        for match in matches:
            try:
                lockfile_path = target_manager.find_single_lockfile(
                    match.path, ecosystem
                )
                if lockfile_path is None:
                    continue
                # Ignore errors here because we assume they are processed later
                deps, _ = parse_lockfile_path(lockfile_path)
                frozen_deps = tuple((dep.package, dep.transitivity) for dep in deps)

                dependency_matches = list(
                    dependencies_range_match_any(depends_on_entries, deps)
                )
                for dep_pat, found_dep in dependency_matches:
                    if found_dep.transitivity == Transitivity(
                        Transitive()
                    ) and transivite_dep_is_also_direct(found_dep.package, frozen_deps):
                        continue
                    reachable_deps.add(
                        (
                            lockfile_path,
                            found_dep.package,
                            found_dep.version,
                            found_dep.transitivity,
                        )
                    )
                    dep_match = DependencyMatch(
                        dependency_pattern=dep_pat,
                        found_dependency=found_dep,
                        lockfile=str(lockfile_path),
                    )
                    match.extra["sca_info"] = ScaInfo(
                        sca_finding_schema=SCA_FINDING_SCHEMA,
                        reachable=True,
                        reachability_rule=rule.should_run_on_semgrep_core,
                        dependency_match=dep_match,
                    )
                    reachable_matches.append(match)
            except SemgrepError as e:
                dep_rule_errors.append(e)
    return (
        reachable_matches,
        dep_rule_errors,
        (lambda p, d: (p, d.package, d.version, d.transitivity) in reachable_deps),
    )



"""
Parser for maven_dep_tree.txt files, generated by maven
Based on the output of this maven plugin https://maven.apache.org/plugins/maven-dependency-plugin/tree-mojo.html
"""
from pathlib import Path
from typing import List
from typing import Optional

from semdep.external.parsy import regex
from semdep.external.parsy import string
from semdep.parsers.util import consume_line
from semdep.parsers.util import mark_line
from semdep.parsers.util import ParserName
from semdep.parsers.util import safe_path_parse
from semgrep.semgrep_interfaces.semgrep_output_v1 import Direct
from semgrep.semgrep_interfaces.semgrep_output_v1 import Ecosystem
from semgrep.semgrep_interfaces.semgrep_output_v1 import FoundDependency
from semgrep.semgrep_interfaces.semgrep_output_v1 import Maven
from semgrep.semgrep_interfaces.semgrep_output_v1 import Transitive
from semgrep.semgrep_interfaces.semgrep_output_v1 import Transitivity

# Annoying to read. In english "6 colon separated strings or 5 colon separated strings"
# Examples:
# org.apache.logging.log4j:log4j-api:jar:0.0.2:compile
# org.springframework.boot:spring-boot-configuration-processor:jar:2.3.4.RELEASE:compile (optional)
# com.google.inject:guice:jar:no_aop:4.2.2:test
dep = regex(
    "([^:\n]+:[^:\n]+):[^:\n]+:[^:\n]+:([^:\n]+):[^:\n]+", flags=0, group=(1, 2)
) | regex("([^:\n]+:[^:\n]+):[^:\n]+:([^:\n]+):[^:\n]+", flags=0, group=(1, 2))


# Examples (these would not appear in this order in a file, they're seperate):
# |  +- org.apache.maven:maven-model:jar:3.8.6:provided

# |  |  \- org.codehaus.plexus:plexus-component-annotations:jar:1.5.5:provided

# +- org.apache.logging.log4j:log4j-api:jar:0.0.2:compile

#    \- net.java.dev.jna:jna:jar:5.11.0:compile

# |     +- org.springframework:spring-aop:jar:5.3.9:compile
tree_line = mark_line(
    regex(r"((\|  )|(   ))*").bind(
        lambda depth: (regex("(\\+- )|(\\\\- )"))
        >> dep.map(
            lambda d: (
                Transitivity(Transitive() if len(depth) // 3 > 0 else Direct()),
                d[0],
                d[1],
            )
        )
        # ignore lines that we don't recognize
        | consume_line
    )
)


pom_tree = (
    consume_line  # First line is the name of the current project, ignore it
    >> string("\n")
    >> tree_line.sep_by(string("\n"))
    << string("\n").optional()
)


def parse_pom_tree(tree_path: Path, _: Optional[Path]) -> List[FoundDependency]:
    deps = safe_path_parse(tree_path, pom_tree, ParserName.pomtree)
    if not deps:
        return []
    output = []
    seen_matches = set()
    for line_number, match in deps:
        if match is None:
            continue

        if match in seen_matches:
            continue
        seen_matches.add(match)

        transitivity, package, version = match
        output.append(
            FoundDependency(
                package=package,
                version=version,
                ecosystem=Ecosystem(Maven()),
                allowed_hashes={},
                transitivity=transitivity,
                line_number=line_number,
            )
        )
    return output



from functools import lru_cache
from pathlib import Path
from typing import Callable
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple

from semdep.parsers.composer import parse_composer_lock
from semdep.parsers.gem import parse_gemfile
from semdep.parsers.go_mod import parse_go_mod
from semdep.parsers.gradle import parse_gradle
from semdep.parsers.package_lock import parse_package_lock
from semdep.parsers.pipfile import parse_pipfile
from semdep.parsers.pnpm import parse_pnpm
from semdep.parsers.poetry import parse_poetry
from semdep.parsers.pom_tree import parse_pom_tree
from semdep.parsers.requirements import parse_requirements
from semdep.parsers.util import DependencyParserError
from semdep.parsers.util import ParserName
from semdep.parsers.yarn import parse_yarn
from semgrep.console import console
from semgrep.error import SemgrepError
from semgrep.semgrep_interfaces.semgrep_output_v1 import Cargo
from semgrep.semgrep_interfaces.semgrep_output_v1 import Ecosystem
from semgrep.semgrep_interfaces.semgrep_output_v1 import FoundDependency
from semgrep.semgrep_interfaces.semgrep_output_v1 import Transitivity
from semgrep.semgrep_interfaces.semgrep_output_v1 import Unknown
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)


def parse_cargo(
    lockfile_text: str, manifest_text: Optional[str]
) -> Generator[FoundDependency, None, None]:
    def parse_dep(s: str) -> FoundDependency:
        lines = s.split("\n")[1:]
        dep = lines[0].split("=")[1].strip()[1:-1]
        version = lines[1].split("=")[1].strip()[1:-1]
        if len(lines) >= 3 and lines[3].startswith("checksum"):
            hash = {"sha256": [lines[3].split("=")[1].strip()[1:-1]]}
        else:
            hash = {}
        return FoundDependency(
            package=dep,
            version=version,
            ecosystem=Ecosystem(Cargo()),
            resolved_url=None,
            allowed_hashes=hash,
            transitivity=Transitivity(Unknown()),
        )

    deps = lockfile_text.split("[[package]]")[1:]
    yield from (parse_dep(dep) for dep in deps)


OLD_LOCKFILE_PARSERS = {
    "cargo.lock": parse_cargo,  # Rust
}

NEW_LOCKFILE_PARSERS: Dict[
    str, Callable[[Path, Optional[Path]], List[FoundDependency]]
] = {
    "requirements.txt": parse_requirements,  # Python
    "requirements3.txt": parse_requirements,  # Python
    "maven_dep_tree.txt": parse_pom_tree,  # Java
    "yarn.lock": parse_yarn,  # JavaScript
    "gradle.lockfile": parse_gradle,  # Java
    "pipfile.lock": parse_pipfile,  # Python
    "package-lock.json": parse_package_lock,  # JavaScript
    "gemfile.lock": parse_gemfile,  # Ruby
    "poetry.lock": parse_poetry,  # Python
    "go.mod": parse_go_mod,  # Go
    "pnpm-lock.yaml": parse_pnpm,  # JavaScript
    "composer.lock": parse_composer_lock,  # PHP
}

LOCKFILE_TO_MANIFEST: Dict[str, Optional[str]] = {
    "Pipfile.lock": "Pipfile",
    "poetry.lock": "pyproject.toml",
    "requirements.txt": "requirements.in",
    "requirements3.txt": "requirements.in",
    "package-lock.json": "package.json",
    "yarn.lock": "package.json",
    "composer.lock": "composer.json",
    "Gemfile.lock": None,
    "go.mod": None,
    "Cargo.lock": None,
    "maven_dep_tree.txt": None,
    "gradle.lockfile": "build.gradle",
    "pnpm-lock.yaml": None,
}


def lockfile_path_to_manifest_path(lockfile_path: Path) -> Optional[Path]:
    """
    Given full lockfile path, return the path to the manifest file if it exists

    Return None if file doesnt exist
    """
    path, lockfile_pattern = lockfile_path.parent, lockfile_path.parts[-1]
    manifest_pattern = LOCKFILE_TO_MANIFEST[lockfile_pattern]

    # some lockfiles don't have a manifest
    if not manifest_pattern:
        return None

    manifest_path = path / manifest_pattern
    if not manifest_path.exists():
        return None

    return manifest_path


@lru_cache(maxsize=1000)
def parse_lockfile_path(
    lockfile_path: Path,
) -> Tuple[List[FoundDependency], Optional[DependencyParserError]]:
    """
    Parse a lockfile and return it as a list of dependency objects

    Also returns Optional DependencyParseError as second return value if there was a problem
    parsing the lockfile
    """
    manifest_path = lockfile_path_to_manifest_path(lockfile_path)
    lockfile_name = lockfile_path.name.lower()
    if lockfile_name in NEW_LOCKFILE_PARSERS:
        parse_lockfile = NEW_LOCKFILE_PARSERS[lockfile_name]

        try:
            return (parse_lockfile(lockfile_path, manifest_path), None)
        except DependencyParserError as e:
            return ([], e)

    if lockfile_name in OLD_LOCKFILE_PARSERS:
        lockfile_text = lockfile_path.read_text()
        if manifest_path:
            manifest_text = manifest_path.read_text()
        else:
            manifest_text = None

        try:
            return (
                list(OLD_LOCKFILE_PARSERS[lockfile_name](lockfile_text, manifest_text)),
                None,
            )
        # Such a general except clause is suspect, but the parsing error could be any number of
        # python errors, since our parsers are just using stdlib string processing functions
        # This will avoid catching dangerous to catch things like KeyboardInterrupt and SystemExit
        except Exception as e:
            console.print(f"Failed to parse {lockfile_path} with exception {e}")
            return (
                [],
                DependencyParserError(lockfile_path, ParserName.cargo, str(e)),
            )
    else:
        raise SemgrepError(f"don't know how to parse this filename: {lockfile_path}")



"""
Parsers for pnpm-lock.yaml files
Based on https://github.com/pnpm/spec/blob/master/lockfile/5.2.md
"""
from pathlib import Path
from typing import List
from typing import Optional

from semdep.external.parsy import regex
from semdep.external.parsy import string
from semdep.external.parsy import string_from
from semdep.parsers.util import mark_line
from semdep.parsers.util import pair
from semdep.parsers.util import ParserName
from semdep.parsers.util import safe_path_parse
from semdep.parsers.util import transitivity
from semdep.parsers.util import upto
from semgrep.semgrep_interfaces.semgrep_output_v1 import Ecosystem
from semgrep.semgrep_interfaces.semgrep_output_v1 import FoundDependency
from semgrep.semgrep_interfaces.semgrep_output_v1 import Npm


consume_line = regex(r"[^\n]*\n")

# Section for the direct dependencies, should look something like
# specifiers:
#     '@types/jsdom': ^1.2.3
#     jsdom: 1.2.3
#     typescript: ~1.2.5

whitespace_not_newline = regex("[^\\S\r\n]+")

package = whitespace_not_newline >> upto(":", consume_other=True)
version_specifier = upto("\n")
dep = pair(
    package, string(" ") >> string_from("~", "^").optional() >> version_specifier
)

direct_dependencies_identifier = whitespace_not_newline.optional() >> string(
    "specifiers:"
)

direct_dependencies = (
    direct_dependencies_identifier
    >> string("\n")
    >> mark_line(dep).sep_by(string("\n").at_least(1))
)


# Section for the transitive dependencies, should look something like
# packages:
#
#     /@foo/bar/1.2.3:
#         resolution: {integrity: sha512-...}
#         engines: {node: '>=1.2.3'}
#         dependencies:
#             '@fizz/buzz': 2.3.4
#
#     /min/4.5.6:
#        resolution: {integrity: sha512-...}
#        engines: {node: '>=1.2.3'}
#        dependencies:
#            'max': 2.3.4
#        transitivePeerDependencies:
#             - mathPackage

packages_identifier = whitespace_not_newline.optional() >> string("packages:")

# "/foo/1.2.3:"
# "/@foo/bar/1.2.3:"
raw_dependency = regex(" */(@.+/.+)/([^:]+)", flags=0, group=(1, 2)) | regex(
    " */(.+)/([^:]+)", flags=0, group=(1, 2)
)

# resolution: {integrity: sha512-...}
# transitivePeerDependencies:
#   - mathPackage
not_used_info = regex("( *-[^\n]*)|( *[^:\n]*:[^\n]*)").sep_by(string("\n"))

full_raw_dependency = mark_line(raw_dependency) << not_used_info

all_dependencies = (
    packages_identifier >> string("\n\n") >> full_raw_dependency.sep_by(string("\n\n"))
)

# if using pnpm workspaces, lockfile will have multiple direct dependencies
# sections. in that case, we'll capture all of them, then flatten the array
# below
direct_dependencies_data = (
    consume_line.until(direct_dependencies_identifier) >> direct_dependencies
).at_least(1)
packages_data = consume_line.until(packages_identifier) >> all_dependencies
all_dependency_data = (
    pair(direct_dependencies_data, packages_data) << consume_line.many()
)


def parse_pnpm(lockfile_path: Path, _: Optional[Path]) -> List[FoundDependency]:
    ret = safe_path_parse(lockfile_path, all_dependency_data, ParserName.pnpm_lock)
    if not ret:
        return []
    direct_deps, all_deps = ret
    if not direct_deps or not all_deps:
        return []

    # direct deps is an array of arrays b/c of workspaces
    direct_deps_flattened = [wd for workspace in direct_deps for wd in workspace]

    # packages that start with @ will look like "'@foo/bar'"
    direct_deps_set = {ps.replace("'", "") for _, (ps, _) in direct_deps_flattened}
    output = []
    for line_number, (package_str, version_str) in all_deps:
        if not package_str or not version_str:
            continue
        output.append(
            FoundDependency(
                package=package_str,
                version=version_str,
                ecosystem=Ecosystem(Npm()),
                transitivity=transitivity(direct_deps_set, [package_str]),
                line_number=line_number,
                allowed_hashes={},
            )
        )
    return output



"""
This module provides an object for printing stuff nicely with rich's formatting features.

The rich docs recommend this usage pattern here:
https://rich.readthedocs.io/en/stable/console.html#console-api

See also the semgrep.terminal module which is an earlier attempt to standardize some output configuration,
but is more low level and doesn't really offload logic to other libraries.
"""
from typing import Any
from typing import Optional

from attrs import frozen
from rich import box
from rich.console import Console
from rich.console import Group
from rich.console import RenderableType
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text


@frozen
class Title:
    text: str
    order: int = 1

    def __rich__(self) -> RenderableType:
        if self.order == 1:
            return Padding(Panel(self.text, expand=False, box=box.SQUARE), (2, 0, 0, 0))
        elif self.order == 2:
            return Padding(Text(self.text.upper(), style="underline"), (1, 0, 0, 2))
        elif self.order == 3:
            return Text("  ⮕ " + self.text.upper())
        else:
            raise ValueError(f"Title order must be 1, 2, or 3, not {self.order}")


class AutoIndentingConsole(Console):
    """
    This custom console keeps track of the last title printed,
    and automatically indents the next line
    according to what level title we're under.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.active_title: Optional[Title] = None
        super().__init__(*args, markup=False, **kwargs)

    @property
    def auto_indent_size(self) -> int:
        if self.active_title is None:
            return 0

        if self.active_title.order == 1:
            return 2
        elif self.active_title.order == 2:
            return 2
        elif self.active_title.order == 3:
            return 5
        else:
            return 0

    def reset_title(self, order: int = 1) -> None:
        self.active_title = Title("", order=order)

    @staticmethod
    def extract_title(current: Optional[RenderableType]) -> Optional[Title]:
        while isinstance(current, Title) or hasattr(current, "renderable"):
            if isinstance(current, Title):
                return current
            current = getattr(current, "renderable", None)

        return None

    def print(self, *args: Any, deindent: int = 0, **kwargs: Any) -> None:
        """Override the default print.

        If what we're printing is a title, we remember it.
        If we remember a recent title,
        indent whatever we print now by how much the most recent title prefers.
        """
        indent = 0

        for arg in args:
            title = self.extract_title(arg)
            if title is not None:
                self.active_title = title
                break
        else:
            indent = self.auto_indent_size

        indent = max(0, indent - deindent)

        super().print(Padding.indent(Group(*args), indent), **kwargs)


MAX_WIDTH = 160

console = AutoIndentingConsole(highlighter=None)
console.width = min(console.width, MAX_WIDTH)


if __name__ == "__main__":
    """Print samples of the above components."""
    console.print("[bold]Semgrep output formatting samples:[/bold]", markup=True)
    console.print(Title("Level 1"))
    console.print("auto-indented text")
    console.print(Title("Level 2", order=2))
    console.print("auto-indented text")
    console.print(Title("Level 3", order=3))
    console.print("auto-indented text")
    console.reset_title()
    console.print("auto-indented text after title reset")
    console.print("bit of markup: [/#]")



"""
Parsers for requirements.txt and requirements.in files
Based on https://pip.pypa.io/en/stable/reference/requirements-file-format/
"""
from pathlib import Path
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

from semdep.external.parsy import string
from semdep.external.parsy import string_from
from semdep.external.parsy import success
from semdep.external.parsy import whitespace
from semdep.parsers import preprocessors
from semdep.parsers.util import consume_line
from semdep.parsers.util import mark_line
from semdep.parsers.util import ParserName
from semdep.parsers.util import safe_path_parse
from semdep.parsers.util import transitivity
from semdep.parsers.util import upto
from semgrep.semgrep_interfaces.semgrep_output_v1 import Ecosystem
from semgrep.semgrep_interfaces.semgrep_output_v1 import FoundDependency
from semgrep.semgrep_interfaces.semgrep_output_v1 import Pypi


whitespace = whitespace | string("\\\n")

# We define a package by its possible delimiters instead of trying to define a grammar for package names
package = upto("=", "<", ">", " ", "[", "\n")

# Packages can have extra packages that we can tell pip to install along with them. We ignore these for now
# Examples:
# package[extra1]
# package [extra1, extra2, extra3]
extras_specifier = string("[") >> upto("]", consume_other=True)

# An operator and a version. If it's not an == with a specific version, we ignore it
# Examples:
# == 1.0.0
# >= 2.3.*
# see version specifier regex here https://github.com/pypa/pip/blob/e69e265cb7b60fb2dacbbb2ab8fa3baaf24bfe4d/src/pip/_vendor/packaging/specifiers.py#LL302
version_specifier = string_from("===", "==", ">=", "<=", ">", "<", "~=", "!=").bind(
    lambda operator: whitespace.optional()
    >> upto(" ", ";", ",", "\n").bind(
        lambda version: success(
            (operator, version) if operator == "==" and "*" not in version else None
        )
    )
)

# Examples:
# space-eqeq == 0.6.1
# eqeq-star [security] == 2.8.*, <= 2.8.1 ; python_version < "2.7"
# extras-and-two-conditions[security] == 2.8.3 ; python_version < "2.7" or sys_platform == 'darwin'
# at-url-with-hash @ https://github.com/urllib3/urllib3/archive/refs/tags/1.26.8.zip#look-this-is-not-a-comment
dep = package.bind(
    lambda package: whitespace.optional()
    >> (
        (string("@") >> consume_line)
        | extras_specifier.optional()
        >> whitespace.optional()
        >> version_specifier.sep_by(string(",") >> whitespace.optional()).bind(
            lambda version_specifiers: upto("\n").optional()
            >> success((package, [x for x in version_specifiers if x]))
        )
    )
)

# Lines can be pip flags, we ignore them
# Examples:
# --index-url https://foo.bar
# -e file:local-file
flag_line = (string("--") | string("-")) >> consume_line

# We're going to preprocess the comments away
comment_line = success(None)

# A whole requirements file, can be requirements.txt or requirements.in
requirements = (
    mark_line(flag_line | dep | consume_line | comment_line)
    .sep_by(string("\n").at_least(1))
    .map(lambda xs: [(l, x) for (l, x) in xs if x])
)


# We preprocess the file to remove comments
# It's much easier to just strip them out than to weave them into parsing


# first remove comments
# https://github.com/pypa/pip/blob/e69e265cb7b60fb2dacbbb2ab8fa3baaf24bfe4d/src/pip/_internal/req/req_file.py#LL45
COMMENT_REGEX = r"(^|\s+)#.*$"


def get_manifest_deps(
    parsed: Optional[List[Tuple[int, Tuple[str, List[Tuple[str, str]]]]]]
) -> Optional[Set[str]]:
    return {package for _, (package, _) in parsed} if parsed else None


def parse_requirements(
    lockfile_path: Path, manifest_path: Optional[Path]
) -> List[FoundDependency]:
    deps = safe_path_parse(
        lockfile_path,
        requirements,
        ParserName.requirements,
        preprocess=preprocessors.CommentRemover(),
    )
    if deps is None:
        return []
    manifest_deps = get_manifest_deps(
        safe_path_parse(manifest_path, requirements, ParserName.requirements)
    )
    output = []
    for line_number, (package, constraints) in deps:
        # A package with no pinned version, skip it
        if len(constraints) != 1:
            continue
        operator, version = constraints[0]
        # This should already be enforced by the parser but we'll be careful
        if operator != "==":
            continue
        output.append(
            FoundDependency(
                package=package,
                version=version,
                ecosystem=Ecosystem(Pypi()),
                allowed_hashes={},
                transitivity=transitivity(manifest_deps, [package]),
                line_number=line_number,
            )
        )
    return output
